from __future__ import annotations
from collections import defaultdict
from typing import List, Union, Optional, TYPE_CHECKING
import numpy as np

if TYPE_CHECKING:
    from clayrs.content_analyzer.content_representation.content import Content
    from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
        CombiningTechnique
    from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict
    from clayrs.recsys.content_based_algorithm.centroid_vector.similarities import Similarity

from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
    Centroid
from clayrs.content_analyzer.ratings_manager.ratings import Interaction
from clayrs.recsys.content_based_algorithm.content_based_algorithm import ContentBasedAlgorithm
from clayrs.recsys.content_based_algorithm.exceptions import NoRatedItems, OnlyNegativeItems, \
    NotPredictionAlg, EmptyUserRatings


class CentroidVector(ContentBasedAlgorithm):
    """
    Class that implements a centroid-like recommender. It first gets the centroid of the items that the user liked.
    Then computes the similarity between the centroid and the item of which the ranking score must be predicted.
    It's a ranking algorithm, so it can't do score prediction

    *   It computes the centroid vector of the features of items *liked by the user*
    *   It computes the similarity between the centroid vector and the items of which the ranking score must be
    predicted

    The items liked by a user are those having a rating higher or equal than a specific **threshold**.
    If the threshold is not specified, the average score of all items liked by the user is used.

    Examples:
        * Interested in only a field representation, CosineSimilarity as similarity,
        $threshold = 3$ (Every item with rating $>= 3$ will be considered as positive)

        >>> from clayrs import recsys as rs
        >>> alg = rs.CentroidVector({"Plot": 0}, rs.CosineSimilarity(), 3)

        * Interested in multiple field representations of the items, CosineSimilarity as similarity,
        $threshold = None$ (Every item with rating $>=$ mean rating of the user will be considered as positive)

        >>> alg = rs.CentroidVector(
        >>>                      item_field={"Plot": [0, "tfidf"],
        >>>                                  "Genre": [0, 1],
        >>>                                  "Director": "doc2vec"},
        >>>                      similarity=rs.CosineSimilarity(),
        >>>                      threshold=None)

        !!! info

            After instantiating the CentroidVector algorithm, pass it in the initialization of
            a CBRS and the use its method to calculate ranking for single user or multiple users:

            Examples:
                 
                >>> cbrs = rs.ContentBasedRS(algorithm=alg, ...)
                >>> cbrs.fit_rank(...)
                >>> # ...

    Args:
        item_field: dict where the key is the name of the field
            that contains the content to use, value is the representation(s) id(s) that will be
            used for the said item. The value of a field can be a string or a list,
            use a list if you want to use multiple representations for a particular field.
        similarity: Kind of similarity to use
        threshold: Threshold for the ratings. If the rating is greater than the threshold, it will be considered
            as positive. If the threshold is not specified, the average score of all items liked by the user is used.
        embedding_combiner: `CombiningTechnique` used when embeddings representation must be used but they are in a
            matrix form instead of a single vector (e.g. when WordEmbedding representations must be used you have one
            vector for each word). By default, the `Centroid` of the rows of the matrix is computed
    """
    __slots__ = ('_similarity', '_emb_combiner', '_centroid', '_positive_rated_list')

    def __init__(self, item_field: dict, similarity: Similarity, threshold: float = None,
                 embedding_combiner: CombiningTechnique = Centroid()):
        super().__init__(item_field, threshold)

        self._similarity = similarity
        self._emb_combiner = embedding_combiner
        self._centroid: Optional[np.ndarray] = None
        self._positive_rated_list: Optional[List] = None

    def process_rated(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict):
        """
        Function that extracts features from positive rated items ONLY!
        The extracted features will be used to fit the algorithm (build the centroid).

        Features extracted will be stored in a private attributes of the class.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents
        """
        # a list since there could be duplicate interaction (eg bootstrap partitioning)
        items_scores_dict = defaultdict(list)
        for interaction in user_ratings:
            items_scores_dict[interaction.item_id].append(interaction.score)

        items_scores_dict = dict(sorted(items_scores_dict.items()))  # sort dictionary based on key for reproducibility

        # Load rated items from the path
        loaded_rated_items: List[Union[Content, None]] = available_loaded_items.get_list([item_id
                                                                                          for item_id in
                                                                                          items_scores_dict.keys()])

        # If threshold wasn't passed in the constructor, then we take the mean rating
        # given by the user as its threshold
        threshold = self.threshold
        if threshold is None:
            threshold = self._calc_mean_user_threshold(user_ratings)

        # we extract feature of each POSITIVE item sorted based on its key: IMPORTANT for reproducibility!!
        # otherwise the matrix we feed to sklearn will have input item in different rows each run!
        positive_rated_list = []
        for item in loaded_rated_items:
            if item is not None:

                score_assigned = map(float, items_scores_dict[item.content_id])

                for score in score_assigned:
                    if score >= threshold:
                        positive_rated_list.append(self.extract_features_item(item))

        if len(user_ratings) == 0:
            raise EmptyUserRatings("The user selected doesn't have any ratings!")

        user_id = user_ratings[0].user_id
        if len(loaded_rated_items) == 0 or (loaded_rated_items.count(None) == len(loaded_rated_items)):
            raise NoRatedItems("User {} - No rated items available locally!".format(user_id))
        if len(positive_rated_list) == 0:
            raise OnlyNegativeItems("User {} - There are only negative items available locally!")

        self._positive_rated_list = positive_rated_list

    def fit(self):
        """
        The fit process for the CentroidVector consists in calculating the centroid of the features
        of the positive items ONLY.

        This method uses extracted features of the positive items stored in a private attribute, so
        `process_rated()` must be called before this method.

        The built centroid will also be stored in a private attribute.
        """
        positive_rated_features_fused = self.fuse_representations(self._positive_rated_list, self._emb_combiner,
                                                                  as_array=True)
        self._centroid = positive_rated_features_fused.mean(axis=0)

        # we delete variable used to fit since will no longer be used
        self._positive_rated_list = None

    def predict(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict,
                filter_list: List[str] = None) -> List[Interaction]:
        """
        CentroidVector is not a score prediction algorithm, calling this method will raise
        the `NotPredictionAlg` exception!

        Raises:
            NotPredictionAlg: exception raised since the CentroidVector algorithm is not a score prediction algorithm
        """
        raise NotPredictionAlg("CentroidVector is not a Score Prediction Algorithm!")

    def rank(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict,
             recs_number: int = None, filter_list: List[str] = None) -> List[Interaction]:
        """
        Rank the top-n recommended items for the user. If the recs_number parameter isn't specified,
        All unrated items for the user will be ranked (or only items in the filter list, if specified).

        One can specify which items must be ranked with the `filter_list` parameter,
        in this case ONLY items in the `filter_list` parameter will be ranked.
        One can also pass items already seen by the user with the filter_list parameter.
        Otherwise, **ALL** unrated items will be ranked.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents
            recs_number: number of the top ranked items to return, if None all ranked items will be returned
            filter_list (list): list of the items to rank, if None all unrated items for the user will be ranked

        Returns:
            List of Interactions object in a descending order w.r.t the 'score' attribute, representing the ranking for
                a single user
        """
        try:
            user_id = user_ratings[0].user_id
        except IndexError:
            raise EmptyUserRatings("The user selected doesn't have any ratings!")

        user_seen_items = set([interaction.item_id for interaction in user_ratings])

        # Load items to predict
        if filter_list is None:
            items_to_predict = available_loaded_items.get_list([item_id for item_id in available_loaded_items
                                                                if item_id not in user_seen_items])
        else:
            items_to_predict = available_loaded_items.get_list(filter_list)

        # Extract features of the items to predict
        id_items_to_predict = []
        features_items_to_predict = []
        for item in items_to_predict:
            if item is not None:
                id_items_to_predict.append(item.content_id)
                features_items_to_predict.append(self.extract_features_item(item))

        if len(id_items_to_predict) > 0:
            # Calculate predictions, they are the similarity of the new items with the centroid vector
            features_fused = self.fuse_representations(features_items_to_predict, self._emb_combiner, as_array=True)
            similarities = [self._similarity.perform(self._centroid, item) for item in features_fused]
        else:
            similarities = []

        # Build the item_score dict (key is item_id, value is rank score predicted)
        # and order the keys in descending order
        item_score_dict = dict(zip(id_items_to_predict, similarities))
        ordered_item_ids = sorted(item_score_dict, key=item_score_dict.get, reverse=True)

        # we only save the top-n items_ids corresponding to top-n recommendations
        # (if recs_number is None ordered_item_ids will contain all item_ids as the original list)
        ordered_item_ids = ordered_item_ids[:recs_number]

        # we construct the output data
        rank_interaction_list = [Interaction(user_id, item_id, item_score_dict[item_id])
                                 for item_id in ordered_item_ids]

        return rank_interaction_list

    def __str__(self):
        return "CentroidVector"

    def __repr__(self):
        return f'CentroidVector(item_field={self.item_field}, ' \
               f'similarity={self._similarity}, ' \
               f'threshold={self.threshold}, ' \
               f'embedding_combiner={self._emb_combiner})'


import unittest
from unittest import TestCase
import pandas as pd
import numpy as np

from clayrs.content_analyzer import Ratings
from clayrs.evaluation.metrics.classification_metrics import Precision, Recall, FMeasure, PrecisionAtK, \
    RPrecision, RecallAtK, FMeasureAtK
from clayrs.evaluation.eval_pipeline_modules.metric_evaluator import Split

pred_only_new_items = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u2', 'u2'],
     'item_id': ['inew1', 'inew2', 'inew3', 'inew4'],
     'score': [650, 600, 500, 650]})
pred_only_new_items = Ratings.from_dataframe(pred_only_new_items)

pred_w_new_items = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u2'],
     'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i4', 'i6', 'i1', 'i8'],
     'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100, 50]})
pred_w_new_items = Ratings.from_dataframe(pred_w_new_items)

pred_only_one_item = pd.DataFrame(
    {'user_id': ['u1', 'u2'],
     'item_id': ['i4', 'i8'],
     'score': [650, 600]})
pred_only_one_item = Ratings.from_dataframe(pred_only_one_item)

pred_i1_i4_missing = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
     'item_id': ['i2', 'i5', 'i6', 'i3', 'i8', 'i9', 'i6', 'i1', 'i8'],
     'score': [600, 400, 300, 220, 100, 50, 200, 100, 50]})
pred_i1_i4_missing = Ratings.from_dataframe(pred_i1_i4_missing)

truth = pd.DataFrame({'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
                      'item_id': ['i1', 'i2', 'i3', 'i4', 'i6', 'i1', 'i8', 'i4'],
                      'score': [3, 2, 3, 1, 2, 4, 3, 3]})
truth = Ratings.from_dataframe(truth)

split_only_new = Split(pred_only_new_items, truth)
split_w_new_items = Split(pred_w_new_items, truth)
split_only_one = Split(pred_only_one_item, truth)
split_missing = Split(pred_i1_i4_missing, truth)


class TestClassificationMetric(TestCase):

    def test_perform_no_relevant_u1(self):
        # any classification metric will work, you just need to change the ratio in the test
        metric_no_relevant_u1 = Precision(relevant_threshold=4)

        result = metric_no_relevant_u1.perform(split_w_new_items)

        result_u1 = result.query('user_id == "u1"')[str(metric_no_relevant_u1)].values[0]
        self.assertTrue(np.isnan(result_u1))

        expected_u2 = 2 / 8
        result_u2 = float(result.query('user_id == "u2"')[str(metric_no_relevant_u1)])
        self.assertAlmostEqual(expected_u2, result_u2)

        expected_sys = expected_u2
        result_sys = float(result.query('user_id == "sys"')[str(metric_no_relevant_u1)])
        self.assertAlmostEqual(expected_sys, result_sys)

        # u1 has no relevant items, u2 = [0, 0, 1, 0]

    def test_no_relevant_items(self):
        truth_no_rel = pd.DataFrame({'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
                                     'item_id': ['i1', 'i2', 'i3', 'i4', 'i6', 'i1', 'i8', 'i4'],
                                     'score': [3, 2, 3, 1, 2, 2, 3, 3]})
        truth_no_rel = Ratings.from_dataframe(truth_no_rel)

        split_no_rel = Split(pred_w_new_items, truth_no_rel)

        # any ClassificationMetric will work
        metric = Precision(relevant_threshold=4)

        with self.assertRaises(ValueError):
            metric.perform(split_no_rel)


class TestPrecision(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric_macro = Precision(relevant_threshold=3, sys_average='macro')
        cls.metric_micro = Precision(relevant_threshold=3, sys_average='micro')
        cls.metric_mean = Precision(sys_average='macro')

    def test_perform_only_new(self):
        metric_macro = self.metric_macro
        metric_micro = self.metric_micro
        metric_mean = self.metric_mean

        result = metric_macro.perform(split_only_new)[str(metric_macro)]
        self.assertEqual(0, all(result))

        result = metric_micro.perform(split_only_new)[str(metric_micro)]
        self.assertEqual(0, all(result))

        result = metric_mean.perform(split_only_new)[str(metric_mean)]
        self.assertEqual(0, all(result))

        # u1 = [0, 0], u2 = [0, 0]

    def test_perform_w_new_items_macro(self):
        metric_macro = self.metric_macro

        result_macro = metric_macro.perform(split_w_new_items)

        expected_u1 = 2 / 8
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 3 / 4
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_w_new_items_micro(self):
        metric_micro = self.metric_micro

        result_micro = metric_micro.perform(split_w_new_items)

        expected_u1 = 2 / 8
        result_micro_u1 = float(result_micro.query('user_id == "u1"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u1, result_micro_u1)

        expected_u2 = 3 / 4
        result_micro_u2 = float(result_micro.query('user_id == "u2"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u2, result_micro_u2)

        expected_micro_sys = (2 + 3) / ((2 + 3) + (6 + 1))
        result_micro_sys = float(result_micro.query('user_id == "sys"')[str(metric_micro)])
        self.assertAlmostEqual(expected_micro_sys, result_micro_sys)

        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_w_new_items_mean(self):
        metric_mean = self.metric_mean

        result_mean = metric_mean.perform(split_w_new_items)

        expected_u1 = 2 / 8
        result_mean_u1 = float(result_mean.query('user_id == "u1"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        expected_u2 = 1 / 4
        result_mean_u2 = float(result_mean.query('user_id == "u2"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result_mean.query('user_id == "sys"')[str(metric_mean)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [0, 0, 1, 0]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_only_one_macro(self):
        metric_macro = self.metric_macro

        result_macro = metric_macro.perform(split_only_one)

        expected_u1 = 0 / 1
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 1 / 1
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # u1 = [0], u2 = [1]

    def test_perform_only_one_micro(self):
        metric_micro = self.metric_micro

        result_micro = metric_micro.perform(split_only_one)

        expected_u1 = 0 / 1
        result_micro_u1 = float(result_micro.query('user_id == "u1"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u1, result_micro_u1)

        expected_u2 = 1 / 1
        result_micro_u2 = float(result_micro.query('user_id == "u2"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u2, result_micro_u2)

        expected_micro_sys = (0 + 1) / ((0 + 1) + (1 + 0))
        result_micro_sys = float(result_micro.query('user_id == "sys"')[str(metric_micro)])
        self.assertAlmostEqual(expected_micro_sys, result_micro_sys)

        # u1 = [0], u2 = [1]

    def test_perform_only_one_mean(self):
        metric_mean = self.metric_mean

        result_mean = metric_mean.perform(split_only_one)

        expected_u1 = 0 / 1
        result_mean_u1 = float(result_mean.query('user_id == "u1"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        expected_u2 = 0 / 1
        result_mean_u2 = float(result_mean.query('user_id == "u2"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result_mean.query('user_id == "sys"')[str(metric_mean)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # u1 = [0], u2 = [0]
        # mean_u1 = 2.2, mean_u2 = 3.33


# Only new tests other than those of TestPrecision, to remove redundancy
class TestPrecisionAtK(TestCase):
    def test_perform(self):
        metric = PrecisionAtK(k=3, relevant_threshold=3)

        result = metric.perform(split_w_new_items)

        expected_u1 = 1 / 3
        result_macro_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 2 / 3
        result_macro_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [1, 0, 1, 1]
        # u1@k = [0, 1, 0], u2@k = [1, 0, 1]

    def test_perform_greater_k(self):
        metric = PrecisionAtK(k=999, relevant_threshold=3)

        result = metric.perform(split_w_new_items)
        expected = Precision(relevant_threshold=3).perform(split_w_new_items)

        result = np.sort(result, axis=0)
        expected = np.sort(expected, axis=0)

        # If k > than number of rows, then it's a regular Precision
        self.assertTrue(np.array_equal(expected, result))

    def test_perform_mean(self):
        metric = PrecisionAtK(k=3)

        result = metric.perform(split_w_new_items)

        expected_u1 = 1 / 3
        result_mean_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        expected_u2 = 1 / 3
        result_mean_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [0, 0, 1, 0]
        # u1@k = [0, 1, 0], u2@k = [0, 0, 1]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_i1_i4_missing(self):
        metric = PrecisionAtK(k=3, relevant_threshold=3)

        result_macro = metric.perform(split_missing)

        expected_u1 = 0 / 3
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 2 / 3
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # u1_total = [0, 0, 0, 1, 0, 0], u2_total = [0, 1, 1]
        # u1@k = [0, 0, 0], u2@k = [0, 1, 1]


# Only new tests other than those of TestPrecision, to remove redundancy
class TestRPrecision(TestCase):
    def test_perform(self):
        metric = RPrecision(relevant_threshold=3)

        result = metric.perform(split_w_new_items)

        expected_u1 = 1 / 2
        result_macro_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 2 / 3
        result_macro_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [1, 0, 1, 1]
        # u1@r = [0, 1], u2@r = [1, 0, 1]

    def test_perform_mean(self):
        metric = RPrecision()

        result = metric.perform(split_w_new_items)

        expected_u1 = 1 / 2
        result_mean_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        expected_u2 = 0 / 1
        result_mean_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [0, 0, 1, 0]
        # u1@r = [0, 1], u2@r = [0]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_i1_i4_missing(self):
        metric = RPrecision(relevant_threshold=3)

        result_macro = metric.perform(split_missing)

        expected_u1 = 0 / 2
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 2 / 3
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # u1_total = [0, 0, 0, 0, 1, 0, 0], u2_total = [0, 0, 1, 1]
        # u1@r = [0, 0], u2@r = [0, 0]


class TestRecall(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric_macro = Recall(relevant_threshold=3, sys_average='macro')
        cls.metric_micro = Recall(relevant_threshold=3, sys_average='micro')
        cls.metric_mean = Recall(sys_average='macro')

    def test_perform_only_new(self):
        metric_macro = self.metric_macro
        metric_micro = self.metric_micro
        metric_mean = self.metric_mean

        result = metric_macro.perform(split_only_new)[str(metric_macro)]
        self.assertEqual(0, all(result))

        result = metric_micro.perform(split_only_new)[str(metric_micro)]
        self.assertEqual(0, all(result))

        result = metric_mean.perform(split_only_new)[str(metric_mean)]
        self.assertEqual(0, all(result))

        # u1 = [0, 0], u2 = [0, 0]

    def test_perform_w_new_items_macro(self):
        metric_macro = self.metric_macro

        result_macro = metric_macro.perform(split_w_new_items)

        expected_u1 = 2 / 2
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 3 / 3
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_w_new_items_micro(self):
        metric_micro = self.metric_micro

        result_micro = metric_micro.perform(split_w_new_items)

        expected_u1 = 2 / 2
        result_micro_u1 = float(result_micro.query('user_id == "u1"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u1, result_micro_u1)

        expected_u2 = 3 / 3
        result_micro_u2 = float(result_micro.query('user_id == "u2"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u2, result_micro_u2)

        expected_micro_sys = (2 + 3) / ((2 + 3) + (0 + 0))
        result_micro_sys = float(result_micro.query('user_id == "sys"')[str(metric_micro)])
        self.assertAlmostEqual(expected_micro_sys, result_micro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_w_new_items_mean(self):
        metric_mean = self.metric_mean

        result_mean = metric_mean.perform(split_w_new_items)

        expected_u1 = 2 / 2
        result_mean_u1 = float(result_mean.query('user_id == "u1"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        expected_u2 = 1 / 1
        result_mean_u2 = float(result_mean.query('user_id == "u2"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result_mean.query('user_id == "sys"')[str(metric_mean)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 1
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [0, 0, 1, 0]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_only_one_macro(self):
        metric_macro = self.metric_macro

        result_macro = metric_macro.perform(split_only_one)

        expected_u1 = 0 / 2
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 1 / 3
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0], u2 = [1]

    def test_perform_only_one_micro(self):
        metric_micro = self.metric_micro

        result_micro = metric_micro.perform(split_only_one)

        expected_u1 = 0 / 2
        result_micro_u1 = float(result_micro.query('user_id == "u1"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u1, result_micro_u1)

        expected_u2 = 1 / 3
        result_micro_u2 = float(result_micro.query('user_id == "u2"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u2, result_micro_u2)

        expected_micro_sys = (0 + 1) / ((0 + 1) + (2 + 2))
        result_micro_sys = float(result_micro.query('user_id == "sys"')[str(metric_micro)])
        self.assertAlmostEqual(expected_micro_sys, result_micro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0], u2 = [1]

    def test_perform_only_one_mean(self):
        metric_mean = self.metric_mean

        result_mean = metric_mean.perform(split_only_one)

        expected_u1 = 0 / 2
        result_mean_u1 = float(result_mean.query('user_id == "u1"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        expected_u2 = 0 / 1
        result_mean_u2 = float(result_mean.query('user_id == "u2"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result_mean.query('user_id == "sys"')[str(metric_mean)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 1
        # u1 = [0], u2 = [0]
        # mean_u1 = 2.2, mean_u2 = 3.33


# Only new tests other than those of TestRecall, to remove redundancy
class TestRecallAtK(TestCase):
    def test_perform(self):
        metric = RecallAtK(k=3, relevant_threshold=3)

        result = metric.perform(split_w_new_items)

        expected_u1 = 1 / 2
        result_macro_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 2 / 3
        result_macro_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [1, 0, 1, 1]
        # u1@k = [0, 1, 0], u2@k = [1, 0, 1]

    def test_perform_greater_k(self):
        metric = RecallAtK(k=999, relevant_threshold=3)

        result = metric.perform(split_w_new_items)
        expected = Recall(relevant_threshold=3).perform(split_w_new_items)

        result = np.sort(result, axis=0)
        expected = np.sort(expected, axis=0)

        # If k > than number of rows, then it's a regular Precision
        self.assertTrue(np.array_equal(expected, result))

    def test_perform_mean(self):
        metric = RecallAtK(k=3)

        result = metric.perform(split_w_new_items)

        expected_u1 = 1 / 2
        result_mean_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        expected_u2 = 1 / 1
        result_mean_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 1
        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [0, 0, 1, 0]
        # u1@k = [0, 1, 0], u2@k = [0, 0, 1]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_i1_i4_missing(self):
        metric = RecallAtK(k=3, relevant_threshold=3)

        result_macro = metric.perform(split_missing)

        expected_u1 = 0 / 2
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        expected_u2 = 2 / 3
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1_total = [0, 0, 0, 1, 0, 0], u2_total = [0, 1, 1]
        # u1@k = [0, 0, 0], u2@k = [0, 1, 1]


class TestFMeasure(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric_macro = FMeasure(relevant_threshold=3, sys_average='macro')
        cls.metric_micro = FMeasure(relevant_threshold=3, sys_average='micro')
        cls.metric_mean = FMeasure(sys_average='macro')

        cls.metric_f2 = FMeasure(beta=2, relevant_threshold=3)
        cls.metric_f3_2 = FMeasure(beta=3.2, relevant_threshold=3)

    @staticmethod
    def fscore(beta: float, prec: float, reca: float):
        beta_2 = beta ** 2
        num = prec * reca
        den = (beta_2 * prec) + reca

        fbeta = (1 + beta_2) * (den and num / den or 0)

        fbeta = fbeta

        return fbeta

    def test_perform_only_new(self):
        metric_macro = self.metric_macro
        metric_micro = self.metric_micro
        metric_mean = self.metric_mean

        result = metric_macro.perform(split_only_new)[str(metric_macro)]
        self.assertEqual(0, all(result))

        result = metric_micro.perform(split_only_new)[str(metric_micro)]
        self.assertEqual(0, all(result))

        result = metric_mean.perform(split_only_new)[str(metric_mean)]
        self.assertEqual(0, all(result))

        # u1 = [0, 0], u2 = [0, 0]

    def test_perform_w_new_items_macro(self):
        metric_macro = self.metric_macro

        result_macro = metric_macro.perform(split_w_new_items)

        prec_u1 = 2 / 8
        reca_u1 = 2 / 2
        expected_u1 = self.fscore(1, prec_u1, reca_u1)
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        prec_u2 = 3 / 4
        reca_u2 = 3 / 3
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_w_new_items_micro(self):
        metric_micro = self.metric_micro

        result_micro = metric_micro.perform(split_w_new_items)

        prec_u1 = 2 / 8
        reca_u1 = 2 / 2
        expected_u1 = self.fscore(1, prec_u1, reca_u1)
        result_micro_u1 = float(result_micro.query('user_id == "u1"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u1, result_micro_u1)

        prec_u2 = 3 / 4
        reca_u2 = 3 / 3
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_micro_u2 = float(result_micro.query('user_id == "u2"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u2, result_micro_u2)

        prec_sys = (2 + 3) / ((2 + 3) + (6 + 1))
        reca_sys = (2 + 3) / ((2 + 3) + (0 + 0))
        expected_micro_sys = self.fscore(1, prec_sys, reca_sys)
        result_micro_sys = float(result_micro.query('user_id == "sys"')[str(metric_micro)])
        self.assertAlmostEqual(expected_micro_sys, result_micro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_w_new_items_mean(self):
        metric_mean = self.metric_mean

        result_mean = metric_mean.perform(split_w_new_items)

        prec_u1 = 2 / 8
        reca_u1 = 2 / 2
        expected_u1 = self.fscore(1, prec_u1, reca_u1)
        result_mean_u1 = float(result_mean.query('user_id == "u1"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        prec_u2 = 1 / 4
        reca_u2 = 1 / 1
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_mean_u2 = float(result_mean.query('user_id == "u2"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result_mean.query('user_id == "sys"')[str(metric_mean)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_only_one_macro(self):
        metric_macro = self.metric_macro

        result_macro = metric_macro.perform(split_only_one)

        prec_u1 = 0 / 1
        reca_u1 = 0 / 2
        expected_u1 = 0
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        prec_u2 = 1 / 1
        reca_u2 = 1 / 3
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0], u2 = [1]

    def test_perform_only_one_micro(self):
        metric_micro = self.metric_micro

        result_micro = metric_micro.perform(split_only_one)

        prec_u1 = 0 / 1
        reca_u1 = 0 / 2
        expected_u1 = 0
        result_micro_u1 = float(result_micro.query('user_id == "u1"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u1, result_micro_u1)

        prec_u2 = 1 / 1
        reca_u2 = 1 / 3
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_micro_u2 = float(result_micro.query('user_id == "u2"')[str(metric_micro)])
        self.assertAlmostEqual(expected_u2, result_micro_u2)

        prec_sys = (0 + 1) / ((0 + 1) + (1 + 0))
        reca_sys = (0 + 1) / ((0 + 1) + (2 + 2))
        expected_micro_sys = self.fscore(1, prec_sys, reca_sys)
        result_micro_sys = float(result_micro.query('user_id == "sys"')[str(metric_micro)])
        self.assertAlmostEqual(expected_micro_sys, result_micro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0], u2 = [1]

    def test_perform_only_one_mean(self):
        metric_mean = self.metric_mean

        result_mean = metric_mean.perform(split_only_one)

        prec_u1 = 0 / 1
        reca_u1 = 0 / 2
        expected_u1 = 0
        result_mean_u1 = float(result_mean.query('user_id == "u1"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        prec_u2 = 0 / 1
        reca_u2 = 0 / 1
        expected_u2 = 0
        result_mean_u2 = float(result_mean.query('user_id == "u2"')[str(metric_mean)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result_mean.query('user_id == "sys"')[str(metric_mean)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 1
        # u1 = [0], u2 = [0]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_f2(self):
        metric_macro = self.metric_f2

        result_macro = metric_macro.perform(split_w_new_items)

        prec_u1 = 2 / 8
        reca_u1 = 2 / 2
        expected_u1 = self.fscore(2, prec_u1, reca_u1)
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        prec_u2 = 3 / 4
        reca_u2 = 3 / 3
        expected_u2 = self.fscore(2, prec_u2, reca_u2)
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_f_real_value(self):
        metric_macro = self.metric_f3_2

        result_macro = metric_macro.perform(split_w_new_items)

        prec_u1 = 2 / 8
        reca_u1 = 2 / 2
        expected_u1 = self.fscore(3.2, prec_u1, reca_u1)
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        prec_u2 = 3 / 4
        reca_u2 = 3 / 3
        expected_u2 = self.fscore(3.2, prec_u2, reca_u2)
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric_macro)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric_macro)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]


# Only new tests other than those of TestFMeasure, to remove redundancy
class TestFMeasureAtK(TestCase):

    @staticmethod
    def fscore(beta: int, prec: float, reca: float):
        beta_2 = beta ** 2
        num = prec * reca
        den = (beta_2 * prec) + reca

        fbeta = (1 + beta_2) * (num / den)

        fbeta = fbeta

        return fbeta

    def test_perform(self):
        metric = FMeasureAtK(k=3, beta=1, relevant_threshold=3)

        result = metric.perform(split_w_new_items)

        prec_u1 = 1 / 3
        reca_u1 = 1 / 2
        expected_u1 = self.fscore(1, prec_u1, reca_u1)
        result_macro_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        prec_u2 = 2 / 3
        reca_u2 = 2 / 3
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_macro_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [1, 0, 1, 1]
        # u1@k = [0, 1, 0], u2@k = [1, 0, 1]

    def test_perform_greater_k(self):
        metric = FMeasureAtK(k=999, relevant_threshold=3)

        result = metric.perform(split_w_new_items)
        expected = FMeasure(relevant_threshold=3).perform(split_w_new_items)

        result = np.sort(result, axis=0)
        expected = np.sort(expected, axis=0)

        # If k > than number of rows, then it's a regular FMeasure
        self.assertTrue(np.array_equal(expected, result))

    def test_perform_mean(self):
        metric = FMeasureAtK(k=3)

        result = metric.perform(split_w_new_items)

        prec_u1 = 1 / 3
        reca_u1 = 1 / 2
        expected_u1 = self.fscore(1, prec_u1, reca_u1)
        result_mean_u1 = float(result.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_mean_u1)

        prec_u2 = 1 / 3
        reca_u2 = 1
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_mean_u2 = float(result.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_mean_u2)

        expected_mean_sys = (expected_u1 + expected_u2) / 2
        result_mean_sys = float(result.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_mean_sys, result_mean_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 1
        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [0, 0, 1, 0]
        # u1@k = [0, 1, 0], u2@k = [0, 0, 1]
        # mean_u1 = 2.2, mean_u2 = 3.33

    def test_perform_i1_i4_missing(self):
        metric = FMeasureAtK(k=3, relevant_threshold=3)

        result_macro = metric.perform(split_missing)

        prec_u1 = 0 / 3
        reca_u1 = 0 / 2
        expected_u1 = 0
        result_macro_u1 = float(result_macro.query('user_id == "u1"')[str(metric)])
        self.assertAlmostEqual(expected_u1, result_macro_u1)

        prec_u2 = 2 / 3
        reca_u2 = 2 / 3
        expected_u2 = self.fscore(1, prec_u2, reca_u2)
        result_macro_u2 = float(result_macro.query('user_id == "u2"')[str(metric)])
        self.assertAlmostEqual(expected_u2, result_macro_u2)

        expected_macro_sys = (expected_u1 + expected_u2) / 2
        result_macro_sys = float(result_macro.query('user_id == "sys"')[str(metric)])
        self.assertAlmostEqual(expected_macro_sys, result_macro_sys)

        # n_real_relevant_u1 = 2, n_real_relevant_u2 = 3
        # u1_total = [0, 0, 0, 1, 0, 0], u2_total = [0, 1, 1]
        # u1@k = [0, 0, 0], u2@k = [0, 1, 1]


if __name__ == "__main__":
    unittest.main()



import unittest
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error

from clayrs.content_analyzer.ratings_manager.ratings import Prediction, Ratings
from clayrs.evaluation.metrics.error_metrics import MSE, MAE, RMSE
from clayrs.evaluation.eval_pipeline_modules.metric_evaluator import Split

pred_only_new_items = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u2', 'u2'],
     'item_id': ['inew1', 'inew2', 'inew3', 'inew4'],
     'score': [4.23, 3.55, 2.12, 4.56]})
pred_only_new_items = Prediction.from_dataframe(pred_only_new_items)

pred_w_new_items = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u2'],
     'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i4', 'i6', 'i1', 'i8'],
     'score': [3.23, 2.33, 1.58, 4.23, 3.32, 4.8, 3.45, 2.56, 4.1, 3.21, 2.8, 1.57]})
pred_w_new_items = Prediction.from_dataframe(pred_w_new_items)

pred_only_one_item = pd.DataFrame(
    {'user_id': ['u1', 'u2'],
     'item_id': ['i4', 'i8'],
     'score': [4.5, 3.66]})
pred_only_one_item = Prediction.from_dataframe(pred_only_one_item)

truth = pd.DataFrame({'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
                      'item_id': ['i1', 'i2', 'i3', 'i4', 'i6', 'i1', 'i8', 'i4'],
                      'score': [3, 2, 3, 1, 2, 4, 2, 3]})
truth = Ratings.from_dataframe(truth)
truth_score_u1 = [3, 2, 3, 1, 2]
truth_score_u2 = [4, 2, 3]

split_only_new = Split(pred_only_new_items, truth)
split_w_new_items = Split(pred_w_new_items, truth)
split_only_one = Split(pred_only_one_item, truth)


class TestMAE(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric = MAE()

    def test_perform_only_new(self):
        metric = self.metric

        result = metric.perform(split_only_new)

        scores = result[str(metric)]

        self.assertTrue(all(pd.isna(scores)))

    def test_perform_w_new_items(self):
        metric = self.metric

        result = metric.perform(split_w_new_items)

        # We exclude predicted score for items not in the truth
        u1_predicted_scores = [2.33, 3.23, 4.8, 1.58, 3.32]
        u1_actual_scores = truth_score_u1

        u1_expected = mean_absolute_error(u1_actual_scores, u1_predicted_scores)
        u1_result = float(result.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected, u1_result)

        # We exclude predicted score for items not in the truth
        u2_predicted_scores = [2.8, 1.57, 4.1]
        u2_actual_scores = truth_score_u2

        u2_expected = mean_absolute_error(u2_actual_scores, u2_predicted_scores)
        u2_result = float(result.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected, u2_result)

        sys_expected = (u1_expected + u2_expected) / 2
        sys_result = float(result.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected, sys_result)

    def test_perform_only_one(self):
        metric = self.metric

        result = metric.perform(split_only_one)

        # We exclude predicted score for items not in the truth
        u1_predicted_scores = [4.5]
        u1_actual_scores = [1]

        u1_expected = mean_absolute_error(u1_actual_scores, u1_predicted_scores)
        u1_result = float(result.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected, u1_result)

        # We exclude predicted score for items not in the truth
        u2_predicted_scores = [3.66]
        u2_actual_scores = [2]

        u2_expected = mean_absolute_error(u2_actual_scores, u2_predicted_scores)
        u2_result = float(result.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected, u2_result)

        sys_expected = (u1_expected + u2_expected) / 2
        sys_result = float(result.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected, sys_result)


class TestMSE(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric = MSE()

    def test_perform_only_new(self):
        metric = self.metric

        result = metric.perform(split_only_new)

        scores = result[str(metric)]

        self.assertTrue(all(pd.isna(scores)))

    def test_perform_w_new_items(self):
        metric = self.metric

        result = metric.perform(split_w_new_items)

        # We exclude predicted score for items not in the truth
        u1_predicted_scores = [2.33, 3.23, 4.8, 1.58, 3.32]
        u1_actual_scores = truth_score_u1

        u1_expected = mean_squared_error(u1_actual_scores, u1_predicted_scores)
        u1_result = float(result.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected, u1_result)

        # We exclude predicted score for items not in the truth
        u2_predicted_scores = [2.8, 1.57, 4.1]
        u2_actual_scores = truth_score_u2

        u2_expected = mean_squared_error(u2_actual_scores, u2_predicted_scores)
        u2_result = float(result.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected, u2_result)

        sys_expected = (u1_expected + u2_expected) / 2
        sys_result = float(result.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected, sys_result)

    def test_perform_only_one(self):
        metric = self.metric

        result = metric.perform(split_only_one)

        # We exclude predicted score for items not in the truth
        u1_predicted_scores = [4.5]
        u1_actual_scores = [1]

        u1_expected = mean_squared_error(u1_actual_scores, u1_predicted_scores)
        u1_result = float(result.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected, u1_result)

        # We exclude predicted score for items not in the truth
        u2_predicted_scores = [3.66]
        u2_actual_scores = [2]

        u2_expected = mean_squared_error(u2_actual_scores, u2_predicted_scores)
        u2_result = float(result.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected, u2_result)

        sys_expected = (u1_expected + u2_expected) / 2
        sys_result = float(result.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected, sys_result)


class TestRMSE(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric = RMSE()

    def test_perform_only_new(self):
        metric = self.metric

        result = metric.perform(split_only_new)

        scores = result[str(metric)]

        self.assertTrue(all(pd.isna(scores)))

    def test_perform_w_new_items(self):
        metric = self.metric

        result = metric.perform(split_w_new_items)

        # We exclude predicted score for items not in the truth
        u1_predicted_scores = [2.33, 3.23, 4.8, 1.58, 3.32]
        u1_actual_scores = truth_score_u1

        u1_expected = mean_squared_error(u1_actual_scores, u1_predicted_scores, squared=False)
        u1_result = float(result.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected, u1_result)

        # We exclude predicted score for items not in the truth
        u2_predicted_scores = [2.8, 1.57, 4.1]
        u2_actual_scores = truth_score_u2

        u2_expected = mean_squared_error(u2_actual_scores, u2_predicted_scores, squared=False)
        u2_result = float(result.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected, u2_result)

        sys_expected = (u1_expected + u2_expected) / 2
        sys_result = float(result.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected, sys_result)

    def test_perform_only_one(self):
        metric = self.metric

        result = metric.perform(split_only_one)

        # We exclude predicted score for items not in the truth
        u1_predicted_scores = [4.5]
        u1_actual_scores = [1]

        u1_expected = mean_squared_error(u1_actual_scores, u1_predicted_scores, squared=False)
        u1_result = float(result.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected, u1_result)

        # We exclude predicted score for items not in the truth
        u2_predicted_scores = [3.66]
        u2_actual_scores = [2]

        u2_expected = mean_squared_error(u2_actual_scores, u2_predicted_scores, squared=False)
        u2_result = float(result.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected, u2_result)

        sys_expected = (u1_expected + u2_expected) / 2
        sys_result = float(result.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected, sys_result)


if __name__ == '__main__':
    unittest.main()


from abc import ABC, abstractmethod
from scipy.spatial.distance import cosine as cosine_distance
import numpy as np


class Similarity(ABC):
    """
    Abstract Class for the various types of similarity
    """
    def __init__(self):
        pass

    @abstractmethod
    def perform(self, v1: np.ndarray, v2: np.ndarray):
        """
        Calculates the similarity between v1 and v2
        """
        raise NotImplementedError


class CosineSimilarity(Similarity):
    """
    Computes cosine similarity
    """
    def __init__(self):
        super().__init__()

    def perform(self, v1: np.ndarray, v2: np.ndarray):
        """
        Calculates the cosine similarity between v1 and v2

        Args:
            v1: first numpy array
            v2: second numpy array
        """

        if not v1.any() or not v2.any():
            return 0
        else:
            # Cosine_distance is defined in the scipy library as 1 - cosine_similarity, so:
            # 1 - cosine_distance = 1 - (1 - cosine_similarity) = cosine_similarity
            return 1 - cosine_distance(v1, v2)

    def __str__(self):
        return "CosineSimilarity"

    def __repr__(self):
        return f"CosineSimilarity()"



from __future__ import annotations
from typing import List, Set, Union, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings

import networkx as nx

from clayrs.recsys.graphs.graph import BipartiteDiGraph, Node
from clayrs.recsys.graphs.graph import UserNode, ItemNode
from clayrs.utils.const import logger
from clayrs.utils.context_managers import get_progbar


class NXBipartiteGraph(BipartiteDiGraph):
    """
    Class that implements a Bipartite graph through networkx library.

    !!! info

        A *Bipartite Graph* is a graph which supports only *User* nodes and *Item* nodes. If you need to model also
        other node categories, consider using a Tripartite Graph or a Full Graph

    It creates a graph from an initial Rating object.

    Consider the following matrix representation of the Rating object
    ```
        +------+-----------+-------+
        | User |   Item    | Score |
        +------+-----------+-------+
        | u1   | Tenet     |     4 |
        | u2   | Inception |     5 |
        | ...  | ...       |   ... |
        +------+-----------+-------+
    ```

    The graph will be created with the following interactions:

    ```
                 4
            u1 -----> Tenet
                 5
            u2 -----> Inception
    ```

    where `u1` and `u2` become *User nodes* and `Tenet` and `Inception` become *Item nodes*,
    with the edge weighted depending on the score given

    If the `link_label` parameter is specified, then each link between users and items will be labeled with the label
    specified (e.g. `link_label='score'`):

    ```
            (4, 'score')
        u1 -------------> Tenet
            (5, 'score')
        u2 -------------> Inception
    ```


    Args:
        source_frame: the initial Ratings object needed to create the graph
        link_label: If specified, each link will be labeled with the given label. Default is None
    """

    def __init__(self, source_frame: Ratings = None, link_label: str = None):

        self._graph = nx.DiGraph()

        if source_frame is not None:
            not_none_dict = {}
            if link_label is not None:
                not_none_dict['label'] = link_label

            with get_progbar(source_frame) as progbar:
                progbar.set_description("Creating User->Item links")

                if len(source_frame.timestamp_column) != 0:
                    edges_with_attributes_gen = ((UserNode(interaction.user_id), ItemNode(interaction.item_id),

                                                  # {**x, **y} merges the dicts x and y
                                                  {**not_none_dict, **{'weight': interaction.score,
                                                                       'timestamp': interaction.timestamp}}
                                                  )
                                                 for interaction in progbar)
                else:
                    edges_with_attributes_gen = ((UserNode(interaction.user_id), ItemNode(interaction.item_id),

                                                  # {**x, **y} merges the dicts x and y
                                                  {**not_none_dict, **{'weight': interaction.score}})
                                                 for interaction in progbar)

                self._graph.add_edges_from(edges_with_attributes_gen)

    @property
    def user_nodes(self) -> Set[UserNode]:
        """
        Returns a set of all *User nodes* in the graph
        """
        return set([node for node in self._graph.nodes if isinstance(node, UserNode)])

    @property
    def item_nodes(self) -> Set[ItemNode]:
        """
        Returns a set of all *Item nodes* in the graph
        """
        return set([node for node in self._graph.nodes if isinstance(node, ItemNode)])

    def add_node(self, node: Union[Node, List[Node]]):
        """
        Adds one or multiple Node objects to the graph.
        Since this is a Bipartite Graph, only `User Node` and `Item Node` can be added!

        No duplicates are allowed, but different category nodes with same id are (e.g. `ItemNode('1')` and
        `UserNode('1')`)

        Args:
            node: Node(s) object(s) that needs to be added to the graph

        Raises:
            ValueError: Exception raised when one of the node to add to the graph is not a User or Item node
        """
        if not isinstance(node, list):
            node = [node]

        if any(not isinstance(n, (UserNode, ItemNode)) for n in node):
            raise ValueError("You can only add UserNodes or ItemNodes to a bipartite graph!")

        self._graph.add_nodes_from(node)

    def add_link(self, start_node: Union[Node, List[Node]], final_node: Union[Node, List[Node]],
                 weight: float = None, label: str = None, timestamp: str = None):
        """
        Creates a link connecting the `start_node` to the `final_node`. If two lists are passed, then the node in
        position $i$ in the `start_node` list will be linked to the node in position $i$ in the `final_node` list.

        If nodes to link do not exist, they will be added automatically to the graph. Please remember that since this is
        a Bipartite Graph, only *User nodes* and *Item nodes* can be added!

        A link can be weighted with the `weight` parameter and labeled with the `label` parameter.
        A timestamp can also be specified via `timestamp` parameter.
        All three are optional parameters, so they are not required

        Args:
            start_node: Single Node object or a list of Node objects. They will be the 'head' of the link, since it's a
                directed graph
            final_node (object): Single Node object or a list Node objects. They will be the 'tail' of the link,
                since it's a directed graph
            weight: weight of the link, default is None (no weight)
            label: label of the link, default is None (no label)
            timestamp: timestamp of the link, default is None (no timestamp)
        """
        if not isinstance(start_node, list):
            start_node = [start_node]

        if not isinstance(final_node, list):
            final_node = [final_node]

        self.add_node(start_node)
        self.add_node(final_node)

        not_none_dict = {}
        if label is not None:
            not_none_dict['label'] = label
        if weight is not None:
            not_none_dict['weight'] = weight
        if timestamp is not None:
            not_none_dict['timestamp'] = timestamp

        self._graph.add_edges_from(zip(start_node, final_node),
                                   **not_none_dict)

    def remove_link(self, start_node: Node, final_node: Node):
        """
        Removes the link connecting the `start_node` to the `final_node`.
        If there's no link between the two nodes, then a warning is printed

        Args:
            start_node: *head* node of the link to remove
            final_node: *tail* node of the link to remove
        """
        try:
            self._graph.remove_edge(start_node, final_node)
        except nx.NetworkXError:
            logger.warning("No link exists between the start node and the final node!\n"
                           "No link will be removed")

    def get_link_data(self, start_node: Node, final_node: Node):
        """
        Get link data such as weight, label, timestamp. between the `start_node` and the `final_node`.
        Returns None if said link doesn't exists

        Remember that this is a directed graph so the result differs if 'start_node' and 'final_node'
        are switched.

        Args:
            start_node: Node object from where the link starts
            final_node: Node object to where the link ends
        """
        return self._graph.get_edge_data(start_node, final_node)

    def get_predecessors(self, node: Node) -> List[Node]:
        """
        Returns a list containing the *predecessors* of the node passed.
        Raises TypeError exception if the node doesn't exists in the graph.

        Taken from networkx library:

        > A predecessor of n is a node m such that there exists a directed
        edge from m to n

        For example:
        ```
        # GRAPH:

        I1 <-- U1
        ↑
        U2
        ```

        ```python
        >>> graph.get_predecessors(ItemNode('I1'))
        [User U1, User U2]
        ```

        Args:
            node: Node for which we want to know the predecessors

        Raises:
            TypeError: Exception raised when the node it's not in the graph
        """
        try:
            return list(self._graph.predecessors(node))
        except nx.NetworkXError:
            raise TypeError("The node specified is not in the graph!")

    def get_successors(self, node: Node) -> List[Node]:
        """
        Returns a list containing the successors of the node passed.
        Returns None if the node doesn't exists in the graph.

        Taken from networkx library:
        > A successor of n is a node m such that there exists a directed
        edge from n to m

        For example:
        ```
        U1 --> I2
        ↓
        I1
        ```

        ```python

        >>> graph.get_successors(UserNode('U1'))
        [Item I1, Item I2]
        ```

        Args:
            node: Node for which we want to know the successors

        Raises:
            TypeError: Exception raised when the node it's not in the graph

        """
        try:
            return list(self._graph.successors(node))
        except nx.NetworkXError:
            raise TypeError("The node specified is not in the graph!")

    def node_exists(self, node: Node) -> bool:
        """
        Returns True if the node passed exists in the graph, False otherwise

        Args:
            node: Node to check whether it's present in the graph or not
        """
        r = self._graph.nodes.get(node)
        return r is not None

    def to_networkx(self) -> nx.DiGraph:
        """
        Returns underlying networkx implementation of the graph
        """
        return self._graph

    def degree_centrality(self) -> Dict:
        """
        Calculate the degree centrality for every node in the graph

        Returns:
            Dictionary containing the degree centrality for each node in the graph
        """
        return nx.degree_centrality(self._graph)

    def closeness_centrality(self) -> Dict:
        """
        Calculate the closeness centrality for every node in the graph

        Returns:
            Dictionary containing the closeness centrality for each node in the graph
        """
        return nx.closeness_centrality(self._graph)

    def dispersion(self) -> Dict:
        """
        Calculate the dispersion for every node in the graph

        Returns:
            Dictionary containing the dispersion computed for each node in the graph
        """
        return nx.dispersion(self._graph)

    def remove_node(self, node_to_remove: Union[Node, List[Node]]):
        """
        Removes one or multiple nodes from the graph.
        If one of the nodes to remove is not present in the graph, it is silently ignored

        Args:
            node_to_remove: Single Node object or a list of Node objects to remove from the graph
        """
        if not isinstance(node_to_remove, list):
            node_to_remove = [node_to_remove]

        self._graph.remove_nodes_from(node_to_remove)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return nx.algorithms.isomorphism.is_isomorphic(self._graph, other._graph)
        else:
            return False

    def __str__(self):
        return "NXBipartiteGraph"

    def __repr__(self):
        return str(self)



from __future__ import annotations
import os.path
import re
from pathlib import Path

import numpy as np
import pyaml


from typing import TYPE_CHECKING

# to avoid circular import. Maybe a little 'hacky', better organization for the future?
# This is almost inevitable though, since Report MUST refer to other modules for type hints
if TYPE_CHECKING:
    from clayrs.content_analyzer.content_analyzer_main import ContentAnalyzer
    from clayrs.content_analyzer import Ratings
    from clayrs.evaluation import EvalModel
    from clayrs.recsys.partitioning import Partitioning
    from clayrs.recsys.recsys import RecSys


class Report:
    """
    Class which will generate a YAML report for the whole experiment (or a part of it) depending on the objects
    passed to the `yaml()` function.

    A report will be generated for each module used (`Content Analyzer`, `RecSys`, `Evaluation`).

    Args:
        output_dir: Path of the folder where reports generated will be saved
        ca_report_filename: Filename of the Content Analyzer report
        rs_report_filename: Filename of the Recsys report
        eva_report_filename: Filename of the evaluation report

    """

    def __init__(self, output_dir: str = '.',
                 ca_report_filename: str = 'ca_report',
                 rs_report_filename: str = 'rs_report',
                 eva_report_filename: str = 'eva_report'):

        self._output_dir = output_dir
        self._ca_report_filename = ca_report_filename
        self._rs_report_filename = rs_report_filename
        self._eva_report_filename = eva_report_filename

    @property
    def output_dir(self):
        return self._output_dir

    @output_dir.setter
    def output_dir(self, path):
        self._output_dir = path

    @property
    def ca_report_filename(self):
        return self._ca_report_filename

    @ca_report_filename.setter
    def ca_report_filename(self, filename: str):
        self._ca_report_filename = filename

    @property
    def rs_report_filename(self):
        return self._rs_report_filename

    @rs_report_filename.setter
    def rs_report_filename(self, filename: str):
        self._rs_report_filename = filename

    @property
    def eva_report_filename(self):
        return self._eva_report_filename

    @eva_report_filename.setter
    def eva_report_filename(self, filename: str):
        self._eva_report_filename = filename

    @staticmethod
    def _extract_arguments(repr_string):

        # this is the regex pattern for matching anything inside balanced '(' ')' parenthesis.
        # It only works if the maximum number of recursion is specified (in this case 25),
        # so basically we are saying that if an attribute of object A is an object B, which has an attribute
        # that is object C and this object has an attribute which is an object ...
        # this will break after the 25th object
        n = 25
        balanced_parenthesis_pattern = r"[^()]*?(?:\(" * n + r"[^()]*?" + r"\)[^()]*?)*?" * n

        # here we match the name of the object: we are matching any \w before the open parenthesis '('
        name = re.match(r"(\w+)(?=\()", repr_string)
        parameters_dict = dict()

        if name is not None:
            name = name.group()

            # We separate in tuples attributes and values.
            # NameObject(a1=1.0, a2=2.0) -> [(a1, 1.0), (a2, 2.0)]
            # we are also matching whole nested objects as attributes, with the max recursion discussed above
            # NameObject(a1=1.0, a2=2.0, a3=NestedObject()) -> [(a1, 1.0), (a2, 2.0), (a3, NestedObject())]
            # we are matching any \w preceded by '(' (the beginning) or by a space (it's a follow up parameter)
            # which has as follow-up character the '=' and a nested object after that or anything else.
            # The last part is the bounding, the match ends when ', \w+=' is found (another parameter= or the ')' (the
            # end)
            tuples_list_args = re.findall(r"(?<=\(|\s)(\w+)=(\w+"+balanced_parenthesis_pattern+r"|.*?)(?=,\s\w+=|\))",
                                          repr_string)

            dict_args_string = dict(tuples_list_args)

            for key, val in dict_args_string.items():

                # Recursive call so that each nested object (if any) will be expanded:
                # {'a3': 'NestedObject(a1=NestedObject())'} -> {'a3': 'NestedObject': {'a1': NestedObject: {}}}
                name_nested, dict_args_nested = Report._extract_arguments(val)

                # if a value can be evaluated we transform it into an object (better serializability) or else
                # we take it as a string
                if name_nested is None:
                    try:
                        parameters_dict[key] = eval(val)
                    except (NameError, SyntaxError):
                        parameters_dict[key] = val
                else:
                    parameters_dict[key] = {name_nested: dict_args_nested}

        return name, parameters_dict

    def _report_ca_module(self, content_analyzer: ContentAnalyzer):

        ca_dict = dict()

        ca_dict['source_file'] = content_analyzer._config.source.representative_name
        ca_dict['id_each_content'] = content_analyzer._config.id

        # EXO REPRESENTATIONS
        exo_list = content_analyzer._config.exogenous_representation_list

        ca_dict['exogenous_representations'] = dict() if len(exo_list) != 0 else None

        for exo_config in exo_list:
            string_exo = repr(exo_config.exogenous_technique)
            name_exo, exo_parameters_dict = self._extract_arguments(string_exo)
            ca_dict['exogenous_representations'][name_exo] = exo_parameters_dict

        # FIELD REPRESENTATIONS
        ca_dict['field_representations'] = dict()

        field_names = content_analyzer._config.get_field_name_list()
        for field_name in field_names:

            field_config_list = content_analyzer._config.get_configs_list(field_name)

            for i, field_config in enumerate(field_config_list):

                ca_dict['field_representations']['{}_{}'.format(field_name, str(i))] = dict()

                single_representation_dict = dict()

                string_technique = repr(field_config.content_technique)

                name_technique, parameter_dict_technique = self._extract_arguments(string_technique)

                single_representation_dict[name_technique] = parameter_dict_technique

                single_representation_dict['preprocessing'] = dict() if len(field_config.preprocessing) != 0 else None
                for preprocessing in field_config.preprocessing:
                    string_preprocessing = repr(preprocessing)

                    name_preprocessing, parameter_dict_preprocessing = self._extract_arguments(string_preprocessing)

                    single_representation_dict['preprocessing'][name_preprocessing] = parameter_dict_preprocessing

                ca_dict['field_representations']['{}_{}'.format(field_name, str(i))] = single_representation_dict

        return ca_dict

    def _report_rs_module(self, original_ratings: Ratings, partitioning_technique: Partitioning, recsys: RecSys):

        # To provide a yaml report for recsys object, a rank or predict method must be called first
        if recsys is not None and recsys._yaml_report is None:
            raise ValueError("You must first call with the rank() or predict() method of the recsys object "
                             "before computing the report!")

        rs_dict = dict()

        if original_ratings is not None:
            rs_dict['interactions'] = dict()

            rs_dict['interactions']['n_users'] = len(set(original_ratings.user_id_column))
            rs_dict['interactions']['n_items'] = len(set(original_ratings.item_id_column))
            rs_dict['interactions']['total_interactions'] = len(original_ratings.user_id_column)

            sparsity_numerator = rs_dict['interactions']['total_interactions']
            sparsity_denominator = rs_dict['interactions']['n_users'] * rs_dict['interactions']['n_items']

            sparsity = 1 - sparsity_numerator / sparsity_denominator

            rs_dict['interactions']['sparsity'] = sparsity
            rs_dict['interactions']['min_score'] = float(np.min(original_ratings.score_column))
            rs_dict['interactions']['max_score'] = float(np.max(original_ratings.score_column))
            rs_dict['interactions']['mean_score'] = float(np.mean(original_ratings.score_column))

        if partitioning_technique is not None:
            rs_dict['partitioning'] = dict()

            partitioning_string = repr(partitioning_technique)
            name_partitioning, parameters_partitioning_dict = self._extract_arguments(partitioning_string)

            rs_dict['partitioning'][name_partitioning] = parameters_partitioning_dict

        if recsys is not None:
            rs_dict['recsys'] = dict()

            recsys_string = repr(recsys)
            name_recsys, _ = self._extract_arguments(recsys_string)

            parameters_recsys_dict = dict()

            name_alg, parameters_alg = self._extract_arguments(repr(recsys.algorithm))

            parameters_recsys_dict['algorithm'] = {name_alg: parameters_alg}

            for key, val in recsys._yaml_report.items():

                val = val if isinstance(val, str) else repr(val)

                # each item of the dict of the recsys may be an object with nested objects as attributes,
                # we check this
                name, parameters_dict = self._extract_arguments(val)

                # in case an element of the dict is an object with nested object, name will not be None.
                # else it's a "base" attribute and we add it as it is to the yaml dict
                if name is not None:
                    parameters_recsys_dict[key] = {name: parameters_dict}
                else:
                    parameters_recsys_dict[key] = val

            rs_dict['recsys'][name_recsys] = parameters_recsys_dict

        return rs_dict

    def _report_eva_module(self, eval_model: EvalModel):

        # To provide a yaml report for eval_model object, the fit() method must be called first
        if eval_model._yaml_report_result is None:
            raise ValueError("You must first call with the fit() method of the eval_model object "
                             "before computing the report!")

        eva_dict = dict()

        eva_dict['n_split'] = len(eval_model.pred_list)

        eva_dict['metrics'] = dict()

        for metric in eval_model.metric_list:
            string_metric = repr(metric)

            name_metric, parameters_metric_dict = self._extract_arguments(string_metric)

            eva_dict['metrics'][name_metric] = parameters_metric_dict

        eva_dict['sys_results'] = eval_model._yaml_report_result

        return eva_dict

    def yaml(self, content_analyzer: ContentAnalyzer = None,
             original_ratings: Ratings = None,
             partitioning_technique: Partitioning = None,
             recsys: RecSys = None,
             eval_model: EvalModel = None):
        """
        Main module responsible of generating the `YAML` reports based on the objects passed to this function:

        * If `content_analyzer` is set, then the report for the Content Analyzer will be produced
        * If one between `original_ratings`, `partitioning_technique`, `recsys` is set, then the report for the recsys
        module will be produced.
        * If `eval_model` is set, then the report for the evaluation module will be produced

        **PLEASE NOTE**: by setting the `recsys` parameter, the last experiment conducted will be documented! If no
        experiment is conducted in the current run, then a `ValueError` exception is raised!

        * Same goes for the `eval_model`

        Examples:

            * Generate a report for the Content Analyzer module
            >>> from clayrs import content_analyzer as ca
            >>> from clayrs import utils as ut
            >>> # movies_ca_config = ...  # user defined configuration
            >>> content_a = ca.ContentAnalyzer(movies_config)
            >>> content_a.fit()  # generate and serialize contents
            >>> ut.Report().yaml(content_analyzer=content_a)  # generate yaml

            * Generate a partial report for the RecSys module
            >>> from clayrs import utils as ut
            >>> from clayrs import recsys as rs
            >>> ratings = ca.Ratings(ca.CSVFile(ratings_path))
            >>> pt = rs.HoldOutPartitioning()
            >>> [train], [test] = pt.split_all(ratings)
            >>> ut.Report().yaml(original_ratings=ratings, partitioning_technique=pt)

            * Generate a full report for the RecSys module and evaluation module
            >>> from clayrs import utils as ut
            >>> from clayrs import recsys as rs
            >>> from clayrs import evaluation as eva
            >>>
            >>> # Generate recommendations
            >>> ratings = ca.Ratings(ca.CSVFile(ratings_path))
            >>> pt = rs.HoldOutPartitioning()
            >>> [train], [test] = pt.split_all(ratings)
            >>> alg = rs.CentroidVector()
            >>> cbrs = rs.ContentBasedRS(alg, train_set=train, items_directory=items_path)
            >>> rank = cbrs.fit_rank(test, n_recs=10)
            >>>
            >>> # Evaluate recommendations and generate report
            >>> em = eva.EvalModel([rank], [test], metric_list=[eva.Precision(), eva.Recall()])
            >>> ut.Report().yaml(original_ratings=ratings,
            >>>                  partitioning_technique=pt,
            >>>                  recsys=cbrs,
            >>>                  eval_model=em)

        Args:
            content_analyzer: `ContentAnalyzer` object used to generate complex representation in the experiment
            original_ratings: `Ratings` object representing the original dataset
            partitioning_technique: `Partitioning` object used to split the original dataset
            recsys: `RecSys` object used to produce recommendations/score predictions. Please note that the latest
                experiment run will be documented. If no experiment is run, then an exception is thrown
            eval_model: `EvalModel` object used to evaluate predictions generated. Please note that the latest
                evaluation run will be documented. If no evaluation is run, then an exception is thrown
        """

        def represent_none(self, _):
            return self.represent_scalar('tag:yaml.org,2002:null', 'null')

        def dump_yaml(output_dir, data):
            with open(output_dir, 'w') as yaml_file:
                pyaml.dump(data, yaml_file, sort_dicts=False, safe=True)

        # None values will be represented as 'null' in yaml file.
        # without this, they will simply be represented as an empty string
        pyaml.add_representer(type(None), represent_none)

        if content_analyzer is not None:
            ca_dict = self._report_ca_module(content_analyzer)

            # create folder if it doesn't exist
            Path(self.output_dir).mkdir(parents=True, exist_ok=True)

            output_dir = os.path.join(self.output_dir, f'{self._ca_report_filename}.yml')
            dump_yaml(output_dir, ca_dict)

        if original_ratings is not None or partitioning_technique is not None or recsys is not None:
            rs_dict = self._report_rs_module(original_ratings, partitioning_technique, recsys)

            # create folder if it doesn't exist
            Path(self.output_dir).mkdir(parents=True, exist_ok=True)

            output_dir = os.path.join(self.output_dir, f'{self._rs_report_filename}.yml')
            dump_yaml(output_dir, rs_dict)

        if eval_model is not None:
            eva_dict = self._report_eva_module(eval_model)

            # create folder if it doesn't exist
            Path(self.output_dir).mkdir(parents=True, exist_ok=True)

            output_dir = os.path.join(self.output_dir, f'{self._eva_report_filename}.yml')
            dump_yaml(output_dir, eva_dict)



from sentence_transformers import SentenceTransformer
import numpy as np

from clayrs.content_analyzer.embeddings.embedding_loader.embedding_loader import SentenceEmbeddingLoader


class Sbert(SentenceEmbeddingLoader):
    """
    Class that produces sentences embeddings using sbert.

    The model will be automatically downloaded if not present locally.

    Args:
        model_name_or_file_path: name of the model to download or path where the model is stored
            locally
    """

    def __init__(self, model_name_or_file_path: str = 'paraphrase-distilroberta-base-v1'):
        super().__init__(model_name_or_file_path)

    def load_model(self):
        try:
            return SentenceTransformer(self.reference)
        except (OSError, AttributeError):
            raise FileNotFoundError

    def get_vector_size(self) -> int:
        return self.model.get_sentence_embedding_dimension()

    def get_embedding(self, sentence: str) -> np.ndarray:
        return self.model.encode(sentence, show_progress_bar=False)

    def get_embedding_token(self, sentence: str) -> np.ndarray:
        raise NotImplementedError("The model chosen can't return token embeddings")

    def __str__(self):
        return "Sbert"

    def __repr__(self):
        return f"Sbert(model_name_or_file_path={self.reference})"



from __future__ import annotations
import itertools
import random
from abc import abstractmethod
from collections import Counter, defaultdict

import pandas as pd
import numpy as np
from typing import Dict, Set, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings
    from clayrs.recsys.partitioning import Split

from clayrs.evaluation.metrics.metrics import Metric
from clayrs.evaluation.utils import get_avg_pop, pop_ratio_by_user, get_item_popularity, get_most_popular_items
from clayrs.evaluation.exceptions import NotEnoughUsers
from clayrs.utils.const import logger


class FairnessMetric(Metric):
    """
    Abstract class that generalize fairness metrics
    """

    @abstractmethod
    def perform(self, split: Split):
        raise NotImplementedError


class GroupFairnessMetric(FairnessMetric):
    """
    Abstract class for fairness metrics based on user groups

    It has some concrete methods useful for group divisions, since every subclass needs to split users into groups.

    Args:
        user_groups: Dict containing group names as keys and percentage of users as value, used to
            split users in groups. Users with more popular items rated are grouped into the first group, users with
            slightly less popular items rated are grouped into the second one, etc.
    """

    def __init__(self, user_groups: Dict[str, float]):
        self.__user_groups = user_groups

    @property
    def user_groups(self):
        return self.__user_groups

    @abstractmethod
    def perform(self, split: Split):
        raise NotImplementedError

    @staticmethod
    def get_avg_pop_by_users(data: Ratings, pop_by_items: Dict, group: Set[str] = None) -> Dict[str, float]:
        r"""
        Get the average popularity for each user in the `data` parameter.

        Average popularity of a single user $u$ is defined as:

        $$
        avg\_pop_u = \frac{\sum_{i \in i_u} pop_i}{|i_u|}
        $$

        Args:
            data: The `Ratings` object that will be used to compute average popularity of each user
            pop_by_items: popularity for each label ('label', 'popularity')
            group: (optional) the set of users (user_id)

        Returns:
            Python dictionary containing as keys each user id and as values the average popularity of each user
        """
        if group is None:
            group = set(data.user_id_column)

        list_by_user = {
            user: [interaction.item_id for interaction in data.get_user_interactions(user)]
            for user in group
        }
        avg_pop_by_users = {
            user: get_avg_pop(list_by_user[user], pop_by_items)
            for user in group
        }

        return avg_pop_by_users

    @staticmethod
    def split_user_in_groups(score_frame: Ratings, groups: Dict[str, float],
                             pop_items: Set[str]) -> Dict[str, Set[str]]:
        r"""
        Users are split into groups based on the *groups* parameter, which contains names of the groups as keys,
        and percentage of how many user must contain a group as values. For example:

            groups = {'popular_users': 0.3, 'medium_popular_users': 0.2, 'low_popular_users': 0.5}

        Every user will be inserted in a group based on how many popular items the user has rated (in relation to the
        percentage of users we specified as value in the dictionary):

        * users with many popular items will be inserted into the first group
        * users with niche items rated will be inserted into one of the last groups.

        In general users are grouped by $Popularity\_ratio$ in a descending order. $Popularity\_ratio$ for a
        single user $u$ is defined as:

        $$
        Popularity\_ratio_u = n\_most\_popular\_items\_rated_u / n\_items\_rated_u
        $$

        The *most popular items* are the first `pop_percentage`% items of all items ordered in a descending order by
        popularity.

        The popularity of an item is defined as the number of times it is rated in the `original_ratings` parameter
        divided by the total number of users in the `original_ratings`.

        Args:
            score_frame: the Ratings object
            groups: each key contains the name of the group and each value contains the
                percentage of the specified group. If the groups don't cover the entire user collection,
                the rest of the users are considered in a 'default_diverse' group
            pop_items: set of most popular *item_id* labels

        Returns:
            A python dictionary containing as keys each group name and as values the set of *user_id* belonging to
                the particular group.
        """
        num_of_users = len(set(score_frame.user_id_column))
        if num_of_users < len(groups):
            raise NotEnoughUsers("You can't split in {} groups {} users! "
                                 "Try reducing number of groups".format(len(groups), num_of_users))

        for percentage_chosen in groups.values():
            if not 0 < percentage_chosen <= 1:
                raise ValueError('Incorrect percentage! Valid percentage range: 0 < percentage <= 1')
        total = sum(groups.values())
        if total > 1:
            raise ValueError("Incorrect percentage! Sum of percentage is > than 1")
        elif total < 1:
            raise ValueError("Sum of percentage is < than 1! Please add another group or redistribute percentages "
                             "among already defined group to reach a total of 1!")

        pop_ratio_by_users = pop_ratio_by_user(score_frame, most_pop_items=pop_items)
        pop_ratio_by_users = sorted(pop_ratio_by_users, key=pop_ratio_by_users.get, reverse=True)

        groups_dict: Dict[str, Set[str]] = {}
        last_index = 0
        percentage = 0.0
        for group_name in groups:
            percentage += groups[group_name]
            group_index = round(num_of_users * percentage)
            if group_index == 0:
                logger.warning('Not enough rows for group {}! It will be discarded'.format(group_name))
            else:
                groups_dict[group_name] = set(pop_ratio_by_users[last_index:group_index])
                last_index = group_index
        return groups_dict


class GiniIndex(FairnessMetric):
    r"""
    The Gini Index metric measures inequality in recommendation lists. It's a system wide metric, so only its
    result it will be returned and not those of every user.
    The metric is calculated as such:

    $$
    Gini_{sys} = \frac{\sum_i(2i - n - 1)x_i}{n\cdot\sum_i x_i}
    $$

    Where:

    - $n$ is the total number of distinct items that are being recommended
    - $x_i$ is the number of times that the item $i$ has been recommended

    A perfectly equal recommender system should recommend every item the same number of times, in which case the Gini
    index would be equal to 0. The more the recsys is "disegual", the more the Gini Index is closer to 1

    If the 'top_n' parameter is specified, then the Gini index will measure inequality considering only the first
    *n* items of every recommendation list of all users

    Args:
        top_n: it's a cutoff parameter, if specified the Gini index will be calculated considering only the first
            'n' items of every recommendation list of all users. Default is None
    """

    def __init__(self, top_n: int = None):
        self.__top_n = top_n

    def __str__(self):
        name = "Gini"
        if self.__top_n:
            name += " - Top {}".format(self.__top_n)

        return name

    def __repr__(self):
        return f'GiniIndex(top_n={self.__top_n})'

    def perform(self, split: Split):
        def gini(x: List):
            """
            Inner method which given a list of values, calculates the gini index

            Args:
                x: list of values of which we want to measure inequality
            """
            # The rest of the code requires numpy arrays.
            x = np.asarray(x)
            sorted_x = np.sort(x)
            n = len(x)
            cumx = np.cumsum(sorted_x, dtype=float)
            # The above formula, with all weights equal to 1 simplifies to:
            return (n + 1 - 2 * np.sum(cumx) / cumx[-1]) / n

        predictions = split.pred

        score_dict = {'user_id': [], str(self): []}

        if self.__top_n is not None:
            predictions = itertools.chain.from_iterable([predictions.get_user_interactions(user_id, self.__top_n)
                                                         for user_id in set(predictions.user_id_column)])

        coun = Counter([prediction.item_id for prediction in predictions])

        result = gini(list(coun.values()))

        score_dict['user_id'].append('sys')
        score_dict[str(self)].append(result)

        return pd.DataFrame(score_dict)


class PredictionCoverage(FairnessMetric):
    r"""
    The Prediction Coverage metric measures in percentage how many distinct items are being recommended in relation
    to all available items. It's a system wide metric, so only its result it will be returned and not those of every
    user.
    The metric is calculated as such:

    $$
    Prediction Coverage_{sys} = (\frac{|I_p|}{|I|})\cdot100
    $$

    Where:

    - $I$ is the set of all available items
    - $I_p$ is the set of recommended items

    The $I$ must be specified through the 'catalog' parameter

    Check the 'Beyond Accuracy: Evaluating Recommender Systems by Coverage and Serendipity' paper for more

    Args:
        catalog: set of item id of the catalog on which the prediction coverage must be computed
    """

    def __init__(self, catalog: Set[str]):
        self.__catalog = set(str(item_id) for item_id in catalog)

    def __str__(self):
        return "PredictionCoverage"

    def __repr__(self):
        return f'PredictionCoverage(catalog={self.__catalog})'

    @property
    def catalog(self):
        return self.__catalog

    def _get_covered(self, pred: Ratings):
        """
        Private function which calculates all recommended items given a catalog of all available items (specified in
        the constructor)

        Args:
            pred: Ratings object containing recommendation lists of all users

        Returns:
            Set of distinct items that have been recommended that also appear in the catalog
        """
        catalog = self.catalog
        pred_items = set(pred.item_id_column)
        return pred_items.intersection(catalog)

    def perform(self, split: Split) -> pd.DataFrame:
        prediction = {'user_id': [], str(self): []}
        catalog = {str(item) for item in self.__catalog}  # cast to string in case user is not careful

        pred = split.pred

        covered_items = self._get_covered(pred)

        percentage = (len(covered_items) / len(catalog)) * 100
        coverage_percentage = np.round(percentage, 2)

        prediction['user_id'].append('sys')
        prediction[str(self)].append(coverage_percentage)

        return pd.DataFrame(prediction)


class CatalogCoverage(PredictionCoverage):
    r"""
    The Catalog Coverage metric measures in percentage how many distinct items are being recommended in relation
    to all available items. It's a system wide metric, so only its result it will be returned and not those of every
    user. It differs from the Prediction Coverage since it allows for different parameters to come into play. If no
    parameter is passed then it's a simple Prediction Coverage.
    The metric is calculated as such:

    $$
    Catalog Coverage_{sys} = (\frac{|\bigcup_{j=1...N}reclist(u_j)|}{|I|})\cdot100
    $$

    Where:

    - $N$ is the total number of users
    - $reclist(u_j)$ is the set of items contained in the recommendation list of user $j$
    - $I$ is the set of all available items

    The $I$ must be specified through the 'catalog' parameter

    The recommendation list of every user ($reclist(u_j)$) can be reduced to the first *n* parameter with the
    top-n parameter, so that catalog coverage is measured considering only the most highest ranked items.

    With the 'k' parameter one could specify the number of users that will be used to calculate catalog coverage:
    k users will be randomly sampled and their recommendation lists will be used. The formula above becomes:

    $$
    Catalog Coverage_{sys} = (\frac{|\bigcup_{j=1...k}reclist(u_j)|}{|I|})\cdot100
    $$

    Where:

    - $k$ is the parameter specified

    Obviously 'k' < N, else simply recommendation lists of all users will be used

    Check the 'Beyond Accuracy: Evaluating Recommender Systems  by Coverage and Serendipity' paper and
    page 13 of the 'Comparison of group recommendation algorithms' paper for more

    Args:
        catalog: set of item id of the catalog on which the prediction coverage must be computed
        top_n: it's a cutoff parameter, if specified the Catalog Coverage will be calculated considering only the first
            'n' items of every recommendation list of all users. Default is None
        k: number of users randomly sampled. If specified, k users will be randomly sampled across all users and only
            their recommendation lists will be used to compute the CatalogCoverage
    """

    def __init__(self, catalog: Set[str], top_n: int = None, k: int = None):
        super().__init__(catalog)
        self.__top_n = top_n
        self.__k = k

    def __str__(self):
        # If none of the parameter is passed, then it's simply a PredictionCoverage
        name = "CatalogCoverage (PredictionCov)"

        if self.__top_n:
            name = "CatalogCoverage"
            name += " - Top {}".format(self.__top_n)
        if self.__k:
            name = "CatalogCoverage"
            name += " - {} sampled users".format(self.__k)

        return name

    def __repr__(self):
        return f'CatalogCoverage(catalog={self.catalog}, top_n={self.__top_n}, k={self.__k})'

    def _get_covered(self, pred: Ratings):
        catalog = self.catalog

        if self.__top_n is not None:
            pred = list(itertools.chain.from_iterable([pred.get_user_interactions(user_id, self.__top_n)
                                                       for user_id in set(pred.user_id_column)]))

        # IF k is passed, then we choose randomly k users and calc catalog coverage
        # based on their predictions. We check that k is < n_user since if it's the equal
        # or it's greater, then all predictions generated for all user must be used
        if self.__k is not None and self.__k < len(pred):
            user_list = list(set([interaction_pred.user_id for interaction_pred in pred]))

            sampling = random.choices(user_list, k=self.__k)
            predicted_items = set([interaction_pred.item_id
                                   for interaction_pred in pred if interaction_pred.user_id in sampling])
            covered_items = predicted_items.intersection(catalog)
        else:
            predicted_items = set([interaction_pred.item_id for interaction_pred in pred])
            covered_items = predicted_items.intersection(catalog)

        return covered_items


class DeltaGap(GroupFairnessMetric):
    r"""
    The Delta GAP (Group Average popularity) metric lets you compare the average popularity "requested" by one or
    multiple groups of users and the average popularity "obtained" with the recommendation given by the recsys.
    It's a system wide metric and results of every group will be returned.

    It is calculated as such:

    $$
    \Delta GAP = \frac{recs_GAP - profile_GAP}{profile_GAP}
    $$

    Users are split into groups based on the *user_groups* parameter, which contains names of the groups as keys,
    and percentage of how many user must contain a group as values. For example:

        user_groups = {'popular_users': 0.3, 'medium_popular_users': 0.2, 'low_popular_users': 0.5}

    Every user will be inserted in a group based on how many popular items the user has rated (in relation to the
    percentage of users we specified as value in the dictionary):

    * users with many popular items will be inserted into the first group
    * users with niche items rated will be inserted into one of the last groups.

    In general users are grouped by $Popularity\_ratio$ in a descending order. $Popularity\_ratio$ for a single user $u$
    is defined as:

    $$
    Popularity\_ratio_u = n\_most\_popular\_items\_rated_u / n\_items\_rated_u
    $$

    The *most popular items* are the first `pop_percentage`% items of all items ordered in a descending order by
    popularity.

    The popularity of an item is defined as the number of times it is rated in the `original_ratings` parameter
    divided by the total number of users in the `original_ratings`.

    It can happen that for a particular user of a group no recommendation are available: in that case it will be skipped
    and it won't be considered in the $\Delta GAP$ computation of its group. In case no user of a group has recs
    available, a warning will be printed and the whole group won't be considered.

    If the 'top_n' parameter is specified, then the $\Delta GAP$ will be calculated considering only the first
    *n* items of every recommendation list of all users

    Args:
        user_groups: Dict containing group names as keys and percentage of users as value, used to
            split users in groups. Users with more popular items rated are grouped into the first group, users with
            slightly less popular items rated are grouped into the second one, etc.
        user_profiles: one or more `Ratings` objects containing interactions of the profile of each user
            (e.g. the **train set**). It should be one for each split to evaluate!
        original_ratings: `Ratings` object containing original interactions of the dataset that will be used to
            compute the popularity of each item (i.e. the number of times it is rated divided by the total number of
            users)
        top_n: it's a cutoff parameter, if specified the Gini index will be calculated considering only their first
            'n' items of every recommendation list of all users. Default is None
        pop_percentage: How many (in percentage) *most popular items* must be considered. Default is 0.2
    """

    def __init__(self, user_groups: Dict[str, float], user_profiles: Union[list, Ratings], original_ratings: Ratings,
                 top_n: int = None, pop_percentage: float = 0.2):
        if not 0 < pop_percentage <= 1:
            raise ValueError('Incorrect percentage! Valid percentage range: 0 < percentage <= 1')

        super().__init__(user_groups)
        self._pop_by_item = get_item_popularity(original_ratings)

        if not isinstance(user_profiles, list):
            user_profiles = [user_profiles]
        self._user_profiles = user_profiles
        self.__top_n = top_n
        self._pop_percentage = pop_percentage

    def __str__(self):
        name = "DeltaGap"
        if self.__top_n:
            name += " - Top {}".format(self.__top_n)
        return name

    # not a complete repr, better understand how to manage cases with 'ratings' repr
    def __repr__(self):
        return f"DeltaGap(user_groups={self.user_groups}, top_n={self.__top_n}, pop_percentage={self._pop_percentage})"

    @staticmethod
    def calculate_gap(group: Set[str], avg_pop_by_users: Dict[str, object]) -> float:
        r"""
        Compute the GAP (Group Average Popularity) formula

        $$
        GAP = \frac{\sum_{u \in U}\cdot \frac{\sum_{i \in i_u} pop_i}{|i_u|}}{|G|}
        $$

        Where:

        - $G$ is the set of users
        - $i_u$ is the set of items rated/recommended by/to user $u$
        - $pop_i$ is the popularity of item i

        Args:
            group: the set of users (user_id)
            avg_pop_by_users: average popularity by user

        Returns:
            score (float): gap score
        """
        total_pop = 0
        for user in group:
            if avg_pop_by_users.get(user):
                total_pop += avg_pop_by_users[user]
        return total_pop / len(group)

    @staticmethod
    def calculate_delta_gap(recs_gap: float, profile_gap: float) -> float:
        """
        Compute the ratio between the recommendation gap and the user profiles gap

        Args:
            recs_gap: recommendation gap
            profile_gap: user profiles gap

        Returns:
            score: delta gap measure
        """
        result = 0
        if profile_gap != 0.0:
            result = (recs_gap - profile_gap) / profile_gap

        return result

    def perform(self, split: Split) -> pd.DataFrame:

        # in order to point to the right `user_profile` set each time the
        # `perform()` method is called, we pop the list but add the `user_profile` set
        # back at the end so that DeltaGap is ready for another evaluation without
        # need to instantiate it again
        split_user_profile = self._user_profiles.pop(0)
        self._user_profiles.append(split_user_profile)

        predictions = split.pred

        if self.__top_n:
            predictions = predictions.take_head_all(self.__top_n)

        most_pop_items = get_most_popular_items(self._pop_by_item, self._pop_percentage)
        splitted_user_groups = self.split_user_in_groups(score_frame=split_user_profile, groups=self.user_groups,
                                                         pop_items=most_pop_items)

        split_result = defaultdict(list)
        split_result['user_id'] = ['sys']

        for group_name in splitted_user_groups:

            # we don't consider users of the group for which we do not have any recommendation
            valid_group = splitted_user_groups[group_name].intersection(predictions.user_id_column)

            if len(valid_group) == 0:
                logger.warning(f"Group {group_name} won't be considered in the DeltaGap since no recs is available "
                               f"for any user of said group!")
                continue

            # Computing avg pop by users recs for delta gap
            avg_pop_by_users_recs = self.get_avg_pop_by_users(predictions, self._pop_by_item, valid_group)
            # Computing avg pop by users profiles for delta gap
            avg_pop_by_users_profiles = self.get_avg_pop_by_users(split_user_profile, self._pop_by_item, valid_group)

            # Computing delta gap for every group
            recs_gap = self.calculate_gap(group=valid_group, avg_pop_by_users=avg_pop_by_users_recs)
            profile_gap = self.calculate_gap(group=valid_group, avg_pop_by_users=avg_pop_by_users_profiles)
            group_delta_gap = self.calculate_delta_gap(recs_gap=recs_gap, profile_gap=profile_gap)

            split_result['{} | {}'.format(str(self), group_name)].append(group_delta_gap)

        return pd.DataFrame(split_result)



from unittest import TestCase
import os

from clayrs.content_analyzer.content_representation.content import SimpleField
from clayrs.content_analyzer.field_content_production_techniques.field_content_production_technique import \
    OriginalData
from clayrs.content_analyzer.raw_information_source import JSONFile, DATFile
from test import dir_test_files

file_path = os.path.join(dir_test_files, "movies_info_reduced.json")


class TestOriginalData(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.file_name = 'test.dat'

    def test_produce_content(self):
        technique = OriginalData()

        data_list = technique.produce_content("Title", [], JSONFile(file_path))

        self.assertEqual(len(data_list), 20)
        self.assertIsInstance(data_list[0], SimpleField)

    def test_produce_content_dtype_specified(self):
        technique = OriginalData(dtype=int)

        source = "50"
        with open(self.file_name, 'w') as f:
            f.write(source)

        result = technique.produce_content("0", [], DATFile(self.file_name))

        self.assertIsInstance(result[0], SimpleField)
        self.assertIsInstance(result[0].value, int)

    def test_produce_content_dtype_cant_convert(self):
        technique = OriginalData(dtype=int)

        source = "cant_convert"
        with open(self.file_name, 'w') as f:
            f.write(source)

        with self.assertRaises(ValueError):
            technique.produce_content("0", [], DATFile(self.file_name))

    def doCleanups(self) -> None:
        if os.path.isfile(self.file_name):
            os.remove(self.file_name)

# DECODE POSSIBLE REPRESENTATION: Not implemented for now
#
# class TestDefaultTechnique(TestCase):
#
#     @classmethod
#     def setUpClass(cls) -> None:
#         cls.technique = DefaultTechnique()
#         cls.file_name = 'test.json'
#
#     def test_produce_content(self):
#
#         data_list = self.technique.produce_content("Title", [], JSONFile(file_path))
#
#         self.assertEqual(len(data_list), 20)
#         self.assertIsInstance(data_list[0], SimpleField)
#
#     def test_produce_content_int(self):
#         source = [{"field": '50'}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         result = self.technique.produce_content("field", [], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, int)
#
#         source = [{"field": 50}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         result = self.technique.produce_content("field", [], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, int)
#
#     def test_produce_content_float(self):
#         source = [{"field": '50.23'}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         result = self.technique.produce_content("field", [], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, float)
#
#         source = [{"field": 50.23}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         result = self.technique.produce_content("field", [], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, float)
#
#     def test_produce_content_list(self):
#         source = [{"field": "['52ciao', '78999stringa']"}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         result = self.technique.produce_content("field", [], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, list)
#
#         source = [{"field": ['52ciao', '78999stringa']}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         result = self.technique.produce_content("field", [], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, list)
#
#     def test_produce_content_string(self):
#         source = [{"field": "50"}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         # If preprocessor are specified, then we are sure the framework should import it as a str
#         result = self.technique.produce_content("field", [NLTK(stopwords_removal=True)], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, str)
#
#         source = [{"field": 50}]
#         with open(self.file_name, 'w') as f:
#             json.dump(source, f)
#
#         # If preprocessor are specified, then we are sure the framework should import it as a str
#         result = self.technique.produce_content("field", [NLTK(stopwords_removal=True)], JSONFile(self.file_name))
#
#         self.assertIsInstance(result[0], SimpleField)
#         self.assertIsInstance(result[0].value, str)
#
#     def doCleanups(self) -> None:
#         if os.path.isfile(self.file_name):
#             os.remove(self.file_name)



import os
from unittest import TestCase

from clayrs.content_analyzer import JSONFile
from clayrs.content_analyzer.exogenous_properties_retrieval import DBPediaMappingTechnique, PropertiesFromDataset, \
    BabelPyEntityLinking
from test import dir_test_files

source_path = os.path.join(dir_test_files, 'test_dbpedia', 'movies_info_reduced.json')


class TestDBPediaMappingTechnique(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.raw_source = JSONFile(source_path)

    def test_get_properties_as_labels(self):

        mapp = DBPediaMappingTechnique('dbo:Film', 'Title')

        # Get properties THAT HAVE VALUE in DBPEDIA ONLY
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value["cinematography"], 'http://dbpedia.org/resource/Thomas_E._Ackerman')
        self.assertIsInstance(prop1.value["starring"], list)
        self.assertNotIn("only_local", prop1.value)

        prop2 = results[1]
        self.assertEqual(prop2.value["film director"], "http://dbpedia.org/resource/Christopher_Nolan")

        # Get all properties in DBPEDIA + Get all properties IN LOCAL SOURCE
        # Local properties will be overwritten by dbpedia values if there's a conflict
        mapp.mode = 'all'
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value["completion date"], "")
        self.assertEqual(prop1.value["cinematography"], "http://dbpedia.org/resource/Thomas_E._Ackerman")
        self.assertIsInstance(prop1.value["starring"], list)
        self.assertEqual(prop1.value["only_local"], "")
        self.assertEqual(prop1.value["Title"], "Jumanji")

        prop2 = results[1]
        self.assertEqual(prop2.value["film director"], "http://dbpedia.org/resource/Christopher_Nolan")
        self.assertEqual(prop2.value["only_local"], "")

        # Get all properties in DBPEDIA ONLY
        mapp.mode = 'all_retrieved'
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value["completion date"], "")
        self.assertEqual(prop1.value["cinematography"], "http://dbpedia.org/resource/Thomas_E._Ackerman")
        self.assertIsInstance(prop1.value["starring"], list)
        self.assertNotIn("only_local", prop1.value)

        prop2 = results[1]
        self.assertEqual(prop2.value["completion date"], "")
        self.assertEqual(prop2.value["film director"], "http://dbpedia.org/resource/Christopher_Nolan")
        self.assertIsInstance(prop2.value["starring"], list)
        self.assertNotIn("only_local", prop2.value)

        # Get all properties in LOCAL with value from DBPEDIA
        mapp.mode = 'original_retrieved'
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value["cinematography"], 'http://dbpedia.org/resource/Thomas_E._Ackerman')
        self.assertEqual(prop1.value["only_local"], '')
        self.assertNotIn('starring', prop1.value)

        prop2 = results[1]
        self.assertEqual(prop2.value["Title"], '')
        self.assertEqual(prop2.value["Budget_source"], '')
        self.assertEqual(prop2.value["only_local"], '')

        # # additional_filter Query
        # mapp = DBPediaMappingTechnique('dbo:Film', 'Title', additional_filters={'dbo:budget': 'Budget_source',
        #                                                                           'dbo:wikiPageID': 'wiki_id'})
        # results = mapp.get_properties(self.raw_source)
        #
        # prop1 = results[0]
        # self.assertEqual(prop1.value["cinematography"], 'http://dbpedia.org/resource/Thomas_E._Ackerman')
        #
        # prop2 = results[1]
        # self.assertEqual(prop2.value["film director"], "http://dbpedia.org/resource/Christopher_Nolan")

    def test_get_properties_as_uri(self):

        mapp = DBPediaMappingTechnique('dbo:Film', 'Title', return_prop_as_uri=True)

        uri_prop_cinematography = "http://dbpedia.org/ontology/cinematography"
        uri_prop_starring = "http://dbpedia.org/ontology/starring"
        uri_prop_completionDate = "http://dbpedia.org/ontology/completionDate"
        uri_prop_director = "http://dbpedia.org/ontology/director"

        # Get properties THAT HAVE VALUE in DBPEDIA ONLY
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value[uri_prop_cinematography], 'http://dbpedia.org/resource/Thomas_E._Ackerman')
        self.assertIsInstance(prop1.value[uri_prop_starring], list)
        self.assertNotIn("only_local", prop1.value)

        prop2 = results[1]
        self.assertEqual(prop2.value[uri_prop_director], "http://dbpedia.org/resource/Christopher_Nolan")

        # Get all properties in DBPEDIA + Get all properties IN LOCAL SOURCE
        # Local properties will be overwritten by dbpedia values if there's a conflict
        mapp.mode = 'all'
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value[uri_prop_completionDate], "")
        self.assertEqual(prop1.value[uri_prop_cinematography], "http://dbpedia.org/resource/Thomas_E._Ackerman")
        self.assertIsInstance(prop1.value[uri_prop_starring], list)
        self.assertEqual(prop1.value["only_local"], "")
        self.assertEqual(prop1.value["Title"], "Jumanji")

        prop2 = results[1]
        self.assertEqual(prop2.value[uri_prop_director], "http://dbpedia.org/resource/Christopher_Nolan")
        self.assertEqual(prop2.value["only_local"], "")

        # Get all properties in DBPEDIA ONLY
        mapp.mode = 'all_retrieved'
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value[uri_prop_completionDate], "")
        self.assertEqual(prop1.value[uri_prop_cinematography], "http://dbpedia.org/resource/Thomas_E._Ackerman")
        self.assertIsInstance(prop1.value[uri_prop_starring], list)
        self.assertNotIn("only_local", prop1.value)

        prop2 = results[1]
        self.assertEqual(prop2.value[uri_prop_completionDate], "")
        self.assertEqual(prop2.value[uri_prop_director], "http://dbpedia.org/resource/Christopher_Nolan")
        self.assertIsInstance(prop2.value[uri_prop_starring], list)
        self.assertNotIn("only_local", prop2.value)

        # Get all properties in LOCAL with value from DBPEDIA
        mapp.mode = 'original_retrieved'
        results = mapp.get_properties(self.raw_source)

        prop1 = results[0]
        self.assertEqual(prop1.value[uri_prop_cinematography], 'http://dbpedia.org/resource/Thomas_E._Ackerman')
        self.assertEqual(prop1.value["only_local"], '')
        self.assertNotIn(uri_prop_starring, prop1.value)

        prop2 = results[1]
        self.assertEqual(prop2.value["Title"], '')
        self.assertEqual(prop2.value["Budget_source"], '')
        self.assertEqual(prop2.value["only_local"], '')

        # # additional_filter Query
        # mapp = DBPediaMappingTechnique('Film', 'EN', 'Title', additional_filters={'budget': 'Budget_source',
        #                                                                           'wikiPageID': 'wiki_id'},
        #                                prop_as_uri=True)
        # results = mapp.get_properties(self.raw_source)
        #
        # prop1 = results[0]
        # self.assertEqual(prop1.value[uri_prop_cinematography], 'http://dbpedia.org/resource/Thomas_E._Ackerman')
        #
        # prop2 = results[1]
        # self.assertEqual(prop2.value[uri_prop_director], "http://dbpedia.org/resource/John_Lasseter")

    def test_entity_doesnt_exists(self):
        with self.assertRaises(ValueError):
            DBPediaMappingTechnique("dbo:not_exists", "Title")


class TestPropertiesFromDataset(TestCase):
    def test_get_properties(self):
        raw_source = JSONFile(source_path)

        # mode is 'only_retrieved_evaluated' so all item with a blank value
        # will be discarded
        expected_1 = {"Title": "Jumanji", "Year": "1995", "Rated": "PG", "Released": "15 Dec 1995",
                      "Budget_source": "6.5E7", "wiki_id": "3700174", "runtime (m)": "104.0"}
        expected_2 = {"Title": "Inception", "Budget_source":  "1.6E8"}
        mapp = PropertiesFromDataset()
        results = mapp.get_properties(raw_source)

        prop1 = results[0]
        prop2 = results[1]
        self.assertEqual(prop1.value, expected_1)
        self.assertEqual(prop2.value, expected_2)

        # mode is 'all' so all item are fine, also those with
        # a blank value
        expected_1 = {"Title": "Jumanji", "Year": "1995", "Rated": "PG", "Released": "15 Dec 1995",
                      "Budget_source": "6.5E7", "cinematography": "", "only_local": "", "wiki_id": "3700174",
                      "runtime (m)": "104.0"}
        expected_2 = {"Title": "Inception", "Budget_source": "1.6E8", "only_local": ""}
        mapp.mode = 'all'
        results = mapp.get_properties(raw_source)

        prop1 = results[0]
        prop2 = results[1]
        self.assertEqual(prop1.value, expected_1)
        self.assertEqual(prop2.value, expected_2)

        # field_name_list with 'title' is passed, and mode is 'only_retrieved_evaluated',
        # so only the field 'title' will be retrieved
        expected_1 = {'Title': 'Jumanji'}
        expected_2 = {'Title': 'Inception'}
        mapp = PropertiesFromDataset(field_name_list=['Title'])
        results = mapp.get_properties(raw_source)

        prop1 = results[0]
        prop2 = results[1]
        self.assertEqual(prop1.value, expected_1)
        self.assertEqual(prop2.value, expected_2)

        # field_name_list with a nonexistent field for a raw content is passed,
        # so nothing will be retrieved for the content with the missing field
        expected_1 = {"wiki_id": "3700174"}
        expected_2 = {}
        mapp = PropertiesFromDataset(field_name_list=['wiki_id'])
        results = mapp.get_properties(raw_source)

        prop1 = results[0]
        prop2 = results[1]
        self.assertEqual(prop1.value, expected_1)
        self.assertEqual(prop2.value, expected_2)

        # field_name_list with a nonexistent field for all raw_contents,
        # so nothing will be retrieved
        expected_1 = {}
        expected_2 = {}
        mapp = PropertiesFromDataset(field_name_list=['ppppp'])
        results = mapp.get_properties(raw_source)

        prop1 = results[0]
        prop2 = results[1]
        self.assertEqual(prop1.value, expected_1)
        self.assertEqual(prop2.value, expected_2)


class TestBabelPyEntityLinking(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.raw_source = JSONFile(source_path)

    def test_get_properties(self):

        result = BabelPyEntityLinking('Title').get_properties(self.raw_source)

        entities_content_1 = result[0].value
        self.assertEqual(len(entities_content_1), 1)
        self.assertIn('Jumanji', entities_content_1)
        self.assertNotEqual(entities_content_1['Jumanji']['babelSynsetID'], '')

        entities_content_2 = result[1].value
        self.assertEqual(len(entities_content_2), 1)
        self.assertIn('Inception', entities_content_2)
        self.assertNotEqual(entities_content_2['Inception']['babelSynsetID'], '')

        entities_content_3 = result[2].value
        self.assertEqual(len(entities_content_3), 1)
        self.assertIn('Demon Island', entities_content_3)
        self.assertNotEqual(entities_content_3['Demon Island']['babelSynsetID'], '')

        entities_content_4 = result[3].value
        self.assertEqual(len(entities_content_4), 1)
        self.assertIn('Léon: The Professional', entities_content_4)
        self.assertNotEqual(entities_content_4['Léon: The Professional']['babelSynsetID'], '')

        entities_content_5 = result[4].value
        self.assertEqual(len(entities_content_5), 0)



from __future__ import annotations
import lzma
import os
import pickle

from clayrs.content_analyzer.content_representation.representation_container import RepresentationContainer
from clayrs.content_analyzer.content_representation.content import Content


def load_content_instance(directory: str, content_id: str, only_field_representations: dict = None) -> Content:
    """
    Loads a serialized content
    Args:
        directory: Path to the directory in which the content is stored
        content_id: ID of the content to load (its filename)
        only_field_representations: Specify exactly which representation to load for the content
            (e.g. {'Plot': 0, 'Genres': 1}). Useful for alleviating memory load

    Returns:
        content (Content)
    """
    try:
        content_filename = os.path.join(directory, '{}.xz'.format(content_id))
        with lzma.open(content_filename, "rb") as content_file:
            content = pickle.load(content_file)

        if only_field_representations is not None:
            smaller_content = Content(content_id)
            field_dict_smaller = {}
            for field, repr_id_list in only_field_representations.items():
                field_dict_smaller[field] = [content.get_field_representation(field, repr_id)
                                             for repr_id in repr_id_list]

            for field, repr_list in field_dict_smaller.items():
                ext_id_list = [id if isinstance(id, str) else None for id in only_field_representations[field]]
                field_repr_container = RepresentationContainer(repr_list, ext_id_list)
                smaller_content.append_field(field, field_repr_container)

            content = smaller_content

    except FileNotFoundError:
        content = None

    return content



from unittest import TestCase

from clayrs.content_analyzer.memory_interfaces import KeywordIndex, SearchIndex


class TestIndexInterface(TestCase):
    def test_keyword_serialize(self):
        index = KeywordIndex("./keyword")
        try:
            index.init_writing()
            index.new_content()
            index.new_field("content_id", "0")
            index.new_field("test1", ["this", "is", "a", "test"])
            index.new_field("test2", ["this", "is", "a", "test", "for", "the", "Text", "Interface"])
            index.serialize_content()
            index.stop_writing()
            self.assertEqual(index.get_field("test1", "0"), ["this", "is", "a", "test"])
            self.assertEqual(index.get_tf_idf("test1", "0")["this"], 0.0)
        finally:
            index.delete()

    def test_search_serialize(self):
        index = SearchIndex("search")
        try:
            index.init_writing()
            index.new_content()
            index.new_field("content_id", "0")
            index.new_field("test1", "This is A test")
            index.new_field("test2", "this is a test for the Text Interface")
            index.serialize_content()
            index.stop_writing()
            self.assertEqual(index.get_field("test1", "0"), "This is A test")
        finally:
            index.delete()

    def test_init_writing(self):
        index1 = SearchIndex("./init_writing")
        index2 = SearchIndex("./init_writing")
        index3 = SearchIndex("./init_writing")

        try:
            index1.init_writing()
            index1.new_content()
            index1.new_field("content_id", "0")
            index1.new_field("init_writing", "test1")
            index1.serialize_content()
            index1.stop_writing()

            # test to check that init_writing with False argument doesn't replace the old index but opens it
            index2.init_writing(False)
            index2.new_content()
            index2.new_field("content_id", "1")
            index2.new_field("init_writing", "test2")
            index2.serialize_content()
            index2.stop_writing()
            self.assertEqual(index2.get_field("init_writing", "0"), "test1")
            self.assertEqual(index2.get_field("init_writing", "1"), "test2")

            # test to check that init_writing with True argument replaces the old index
            index3.init_writing(True)
            index3.new_content()
            index3.new_field("content_id", "0")
            index3.new_field("init_writing", "test3")
            index3.serialize_content()
            index3.stop_writing()
            self.assertEqual(index3.get_field("init_writing", "0"), "test3")
            with self.assertRaises(IndexError):
                index3.get_field("init_writing", "1")
        finally:
            index1.delete()
            index2.delete()
            index3.delete()

    def test_get_field(self):
        index = KeywordIndex("./keyword")

        try:
            index.init_writing()
            index.new_content()
            index.new_field("content_id", "0")
            index.new_field("test1", ["this", "is", "a", "test"])
            index.new_field("test2", ["this", "is", "a", "test", "for", "the", "Text", "Interface"])
            index.serialize_content()
            index.stop_writing()

            # test for using get_field with content's id
            result = index.get_field("test1", "0")
            self.assertIsInstance(result, list)
            self.assertEqual(result, ["this", "is", "a", "test"])

            # test for using get_field with content's position in the index
            result = index.get_field("test1", 0)
            self.assertIsInstance(result, list)
            self.assertEqual(result, ["this", "is", "a", "test"])
        finally:
            index.delete()

    def test_get_tfidf(self):
        index = KeywordIndex("./keyword")
        try:
            index.init_writing()
            index.new_content()
            index.new_field("content_id", "0")
            index.new_field("test1", ["this", "is", "a", "test"])
            index.new_field("test2", ["this", "is", "a", "test", "for", "the", "Text", "Interface"])
            index.serialize_content()
            index.stop_writing()

            # test for using get_tf_idf with content's id
            result = index.get_tf_idf("test1", "0")
            self.assertIsInstance(result, dict)
            self.assertEqual(result["this"], 0.0)

            # test for using get_td_idf with content's position in the index
            result = index.get_tf_idf("test1", 0)
            self.assertIsInstance(result, dict)
            self.assertEqual(result["this"], 0.0)
        finally:
            index.delete()

    def test_query(self):
        index = SearchIndex("testing_query")
        try:
            index.init_writing()
            index.new_content()
            index.new_field("content_id", "0")
            index.new_field("test1", "this is a test for the query on the index")
            index.new_field("test2", "this is the second field")
            index.serialize_content()
            index.new_content()
            index.new_field("content_id", "1")
            index.new_field("test1", "field")
            index.serialize_content()
            index.new_content()
            index.new_field("content_id", "2")
            index.new_field("test1", "query on the index")
            index.serialize_content()
            index.stop_writing()

            # test for querying the index
            result = index.query("test1:(query on the index)", 2, ["2"], ["0", "1"], True)
            self.assertEqual(len(result), 1)
            self.assertEqual(result["0"]["item"]["test1"], "this is a test for the query on the index")
        finally:
            index.delete()




import unittest
from unittest import TestCase
import pandas as pd
import numpy as np

from clayrs.content_analyzer import Ratings
from clayrs.recsys.methodology import TestRatingsMethodology, TestItemsMethodology, \
    TrainingItemsMethodology, AllItemsMethodology

train1 = pd.DataFrame.from_records([
    ("001", "tt0112281", 3.5, "54654675"),
    ("001", "tt0112302", 4.5, "54654675"),
    ("002", "tt0112346", 4, "54654675"),
    ("003", "tt0112453", 2, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])

test1 = pd.DataFrame.from_records([
    ("001", "tt0112641", 2, "54654675"),
    ("001", "tt0112760", 1, "54654675"),
    ("002", "tt0112641", 3, "54654675"),
    ("002", "tt0112896", 2, "54654675"),
    ("003", "tt0113041", 3, "54654675"),
    ("003", "tt0112281", 5, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])

train2 = pd.DataFrame.from_records([
    ("001", "tt0112641", 2, "54654675"),
    ("001", "tt0112760", 1, "54654675"),
    ("002", "tt0112641", 3, "54654675"),
    ("002", "tt0112896", 2, "54654675"),
    ("003", "tt0113041", 3, "54654675"),
    ("003", "tt0112281", 5, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])

test2 = pd.DataFrame.from_records([
    ("001", "tt0112281", 3.5, "54654675"),
    ("001", "tt0112302", 4.5, "54654675"),
    ("002", "tt0112346", 4, "54654675"),
    ("003", "tt0112453", 2, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])

train1_rat = Ratings.from_dataframe(train1)
test1_rat = Ratings.from_dataframe(test1)
train2_rat = Ratings.from_dataframe(train2)
test2_rat = Ratings.from_dataframe(test2)


def _unpack_result_w_iter(result_list_w_iter):
    result_list = []
    for result_w_iter in result_list_w_iter:
        result_unpacked = {}
        for user, items_iterator in result_w_iter.items():
            result_unpacked[user] = set(items_iterator)

        result_list.append(result_unpacked)

    return result_list


class TestTestRatingsMethodology(TestCase):

    def test_filter_all(self):

        result_list = [TestRatingsMethodology().filter_all(train1_rat, test1_rat),
                       TestRatingsMethodology().filter_all(train2_rat, test2_rat)]

        # for every user get the items in its test_set1
        expected_list = [test1[['from_id', 'to_id']], test2[['from_id', 'to_id']]]

        self.assertTrue(len(expected_list), len(result_list))

        for expected, result in zip(expected_list, result_list):
            expected = np.array(expected)
            expected.sort(axis=0)

            result = np.array(result)
            result.sort(axis=0)

            self.assertTrue(np.array_equal(expected, result))

    def test_filter_all_only_greater_eq(self):

        result_list = [TestRatingsMethodology(only_greater_eq=3).filter_all(train1_rat, test1_rat),
                       TestRatingsMethodology(only_greater_eq=3).filter_all(train2_rat, test2_rat)]

        # for every user get the items in its test_set1 with score >= 3
        expected_split_1 = pd.DataFrame({
            'from_id': ['002',
                        '003', '003'],
            'to_id': ["tt0112641",
                      "tt0113041", "tt0112281"]
        })

        # for every user get the items in its test_set2 with score >= 3
        expected_split_2 = pd.DataFrame({
            'from_id': ['001', '001',
                        '002'],
            'to_id': ["tt0112281", "tt0112302",
                      "tt0112346"]
        })

        expected_list = [expected_split_1, expected_split_2]

        self.assertTrue(len(expected_list), len(result_list))

        for expected, result in zip(expected_list, result_list):
            expected = np.array(expected)
            expected.sort(axis=0)

            result = np.array(result)
            result.sort(axis=0)

            self.assertTrue(np.array_equal(expected, result))

    def test_result_as_dict_iter(self):
        result_list_iter = [TestRatingsMethodology().filter_all(train1_rat, test1_rat, result_as_iter_dict=True),
                            TestRatingsMethodology().filter_all(train2_rat, test2_rat, result_as_iter_dict=True)]

        # for every user get the items in its test_set1
        expected_list = [{'001': {"tt0112641", "tt0112760"},
                          '002': {"tt0112641", "tt0112896"},
                          '003': {"tt0113041", "tt0112281"}},
                         {'001': {"tt0112281", "tt0112302"},
                          '002': {"tt0112346"},
                          '003': {"tt0112453"}}
                         ]

        result_list = _unpack_result_w_iter(result_list_iter)

        self.assertTrue(len(expected_list), len(result_list))

        self.assertCountEqual(result_list, expected_list)


# poor choice of words sadly
class TestTestItemsMethodology(TestCase):

    def test_filter_all(self):

        result_list = [TestItemsMethodology().filter_all(train1_rat, test1_rat),
                       TestItemsMethodology().filter_all(train2_rat, test2_rat)]

        # for every user get the all items present in test_set1 except the items
        # present in the training_set1 of the user
        expected_split_1 = pd.DataFrame({
            'from_id': ['001', '001', '001', '001',
                        '002', '002', '002', '002', '002',
                        '003', '003', '003', '003', '003'],
            'to_id': ["tt0112641", "tt0112760", "tt0112896", "tt0113041",
                      "tt0112641", "tt0112760", "tt0112896", "tt0113041", "tt0112281",
                      "tt0112641", "tt0112760", "tt0112896", "tt0113041", "tt0112281"]
        })

        # for every user get the all items present in test_set2 except the items
        # present in the training_set2 of the user
        expected_split_2 = pd.DataFrame({
            'from_id': ['001', '001', '001', '001',
                        '002', '002', '002', '002',
                        '003', '003', '003'],
            'to_id': ["tt0112281", "tt0112302", "tt0112346", "tt0112453",
                      "tt0112281", "tt0112302", "tt0112346", "tt0112453",
                      "tt0112302", "tt0112346", "tt0112453"]
        })

        expected_list = [expected_split_1, expected_split_2]

        self.assertTrue(len(expected_list), len(result_list))

        for expected, result in zip(expected_list, result_list):
            expected = np.array(expected)
            expected.sort(axis=0)

            result = np.array(result)
            result.sort(axis=0)

            self.assertTrue(np.array_equal(expected, result))

    def test_filter_all_only_greater_eq(self):
        result_list = [TestItemsMethodology(only_greater_eq=3).filter_all(train1_rat, test1_rat),
                       TestItemsMethodology(only_greater_eq=3).filter_all(train2_rat, test2_rat)]

        # for every user get the all items present in test_set1 with score >= 3 except the items
        # present in the training_set1 of the user
        expected_split_1 = pd.DataFrame({
            'from_id': ['001', '001',
                        '002', '002', '002',
                        '003', '003', '003'],
            'to_id': ["tt0112641", "tt0113041",
                      "tt0112641", "tt0113041", "tt0112281",
                      "tt0112641", "tt0113041", "tt0112281"]
        })

        # for every user get the all items present in test_set2 with score >= 3 except the items
        # present in the training_set2 of the user
        expected_split_2 = pd.DataFrame({
            'from_id': ['001', '001', '001',
                        '002', '002', '002',
                        '003', '003'],
            'to_id': ["tt0112281", "tt0112302", "tt0112346",
                      "tt0112281", "tt0112302", "tt0112346",
                      "tt0112302", "tt0112346"]
        })

        expected_list = [expected_split_1, expected_split_2]

        self.assertTrue(len(expected_list), len(result_list))

        for expected, result in zip(expected_list, result_list):
            expected = np.array(expected)
            expected.sort(axis=0)

            result = np.array(result)
            result.sort(axis=0)

            self.assertTrue(np.array_equal(expected, result))

    def test_result_as_dict(self):
        result_list_iter = [TestItemsMethodology().filter_all(train1_rat, test1_rat, result_as_iter_dict=True),
                            TestItemsMethodology().filter_all(train2_rat, test2_rat, result_as_iter_dict=True)]

        expected_split_1 = {
            '001': ["tt0112641", "tt0112760", "tt0112896", "tt0113041"],
            '002': ["tt0112641", "tt0112760", "tt0112896", "tt0113041", "tt0112281"],
            '003': ["tt0112641", "tt0112760", "tt0112896", "tt0113041", "tt0112281"]
        }

        expected_split_2 = {
            '001': ["tt0112281", "tt0112302", "tt0112346", "tt0112453"],
            '002': ["tt0112281", "tt0112302", "tt0112346", "tt0112453"],
            '003': ["tt0112302", "tt0112346", "tt0112453"]
        }

        expected_list = [expected_split_1, expected_split_2]
        result_list = _unpack_result_w_iter(result_list_iter)

        self.assertTrue(len(expected_list), len(result_list))

        self.assertCountEqual(expected_list[0], result_list[0])
        self.assertCountEqual(expected_list[1], result_list[1])


class TestTrainingItemsMethodology(TestCase):

    def test_filter_all(self):

        result_list = [TrainingItemsMethodology().filter_all(train1_rat, test1_rat),
                       TrainingItemsMethodology().filter_all(train2_rat, test2_rat)]

        # for every user get the all items present in training_set1 except the items
        # present in the training_set1 of the user
        expected_split_1 = pd.DataFrame({
            'from_id': ['001', '001',
                        '002', '002', '002',
                        '003', '003', '003'],
            'to_id': ["tt0112346", "tt0112453",
                      "tt0112281", "tt0112302", "tt0112453",
                      "tt0112281", "tt0112302", "tt0112346"]
        })

        # for every user get the all items present in training_set2 except the items
        # present in the training_set2 of the user
        expected_split_2 = pd.DataFrame({
            'from_id': ['001', '001', '001',
                        '002', '002', '002',
                        '003', '003', '003'],
            'to_id': ["tt0112896", "tt0113041", "tt0112281",
                      "tt0112760", "tt0113041", "tt0112281",
                      "tt0112641", "tt0112760", "tt0112896"]
        })

        expected_list = [expected_split_1, expected_split_2]

        self.assertTrue(len(expected_list), len(result_list))

        for expected, result in zip(expected_list, result_list):
            expected = np.array(expected)
            expected.sort(axis=0)

            result = np.array(result)
            result.sort(axis=0)

            self.assertTrue(np.array_equal(expected, result))

    def test_filter_all_only_greater_eq(self):
        result_list = [TrainingItemsMethodology(only_greater_eq=3).filter_all(train1_rat, test1_rat),
                       TrainingItemsMethodology(only_greater_eq=3).filter_all(train2_rat, test2_rat)]

        # for every user get the all items present in training_set1 with score >= 3 except the items
        # present in the training_set1 of the user
        expected_split_1 = pd.DataFrame({
            'from_id': ['001',
                        '002', '002',
                        '003', '003', '003'],
            'to_id': ["tt0112346",
                      "tt0112281", "tt0112302",
                      "tt0112281", "tt0112302", "tt0112346"]
        })

        # for every user get the all items present in training_set2 with score >= 3 except the items
        # present in the training_set2 of the user
        expected_split_2 = pd.DataFrame({
            'from_id': ['001', '001',
                        '002', '002',
                        '003'],
            'to_id': ["tt0113041", "tt0112281",
                      "tt0113041", "tt0112281",
                      "tt0112641"]
        })

        expected_list = [expected_split_1, expected_split_2]

        self.assertTrue(len(expected_list), len(result_list))

        for expected, result in zip(expected_list, result_list):
            expected = np.array(expected)
            expected.sort(axis=0)

            result = np.array(result)
            result.sort(axis=0)

            self.assertTrue(np.array_equal(expected, result))

    def test_result_as_dict(self):
        result_list_iter = [TrainingItemsMethodology().filter_all(train1_rat, test1_rat, result_as_iter_dict=True),
                            TrainingItemsMethodology().filter_all(train2_rat, test2_rat, result_as_iter_dict=True)]

        expected_split_1 = {
            '001': ["tt0112346", "tt0112453"],
            '002': ["tt0112281", "tt0112302", "tt0112453"],
            '003': ["tt0112281", "tt0112302", "tt0112346"]
        }

        expected_split_2 = {
            '001': ["tt0112896", "tt0113041", "tt0112281"],
            '002': ["tt0112760", "tt0113041", "tt0112281"],
            '003': ["tt0112641", "tt0112760", "tt0112896"]
        }

        expected_list = [expected_split_1, expected_split_2]
        result_list = _unpack_result_w_iter(result_list_iter)

        self.assertTrue(len(expected_list), len(result_list))

        self.assertCountEqual(expected_list[0], result_list[0])
        self.assertCountEqual(expected_list[1], result_list[1])


class TestAllItemsMethodology(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.all_items = ["tt0112281",
                         "tt0112302",
                         "tt0112346",
                         "tt0112453",
                         "iall1",
                         "iall2",
                         "iall3",
                         "iall4"]

    def test_filter_all(self):
        result_list = [AllItemsMethodology(set(self.all_items)).filter_all(train1_rat, test1_rat),
                       AllItemsMethodology(set(self.all_items)).filter_all(train2_rat, test2_rat)]

        expected_split_1 = pd.DataFrame({
            'from_id': ["001", "001", "001", "001", "001", "001",
                        "002", "002", "002", "002", "002", "002", "002",
                        "003", "003", "003", "003", "003", "003", "003"],
            'to_id': ["tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4",
                      "tt0112281", "tt0112302", "tt0112453", "iall1", "iall2", "iall3", "iall4",
                      "tt0112281", "tt0112302", "tt0112346", "iall1", "iall2", "iall3", "iall4", ]
        })

        expected_split_2 = pd.DataFrame({
            'from_id': ["001", "001", "001", "001", "001", "001", "001", "001",
                        "002", "002", "002", "002", "002", "002", "002", "002",
                        "003", "003", "003", "003", "003", "003", "003"],
            'to_id': ["tt0112281", "tt0112302", "tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4",
                      "tt0112281", "tt0112302", "tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4",
                      "tt0112302", "tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4"]
        })

        expected_list = [expected_split_1, expected_split_2]

        self.assertTrue(len(expected_list), len(result_list))

        for expected, result in zip(expected_list, result_list):
            expected = np.array(expected)
            expected.sort(axis=0)

            result = np.array(result)
            result.sort(axis=0)

            self.assertTrue(np.array_equal(expected, result))

    def test_result_as_dict(self):
        result_list_iter = [AllItemsMethodology(set(self.all_items)).filter_all(train1_rat, test1_rat,
                                                                                result_as_iter_dict=True),
                            AllItemsMethodology(set(self.all_items)).filter_all(train2_rat, test2_rat,
                                                                                result_as_iter_dict=True)]

        expected_split_1 = {
            '001': ["tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4"],
            '002': ["tt0112281", "tt0112302", "tt0112453", "iall1", "iall2", "iall3", "iall4"],
            '003': ["tt0112281", "tt0112302", "tt0112346", "iall1", "iall2", "iall3", "iall4"]
        }

        expected_split_2 = {
            '001': ["tt0112281", "tt0112302", "tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4", ],
            '002': ["tt0112281", "tt0112302", "tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4", ],
            '003': ["tt0112302", "tt0112346", "tt0112453", "iall1", "iall2", "iall3", "iall4"]
        }

        expected_list = [expected_split_1, expected_split_2]
        result_list = _unpack_result_w_iter(result_list_iter)

        self.assertTrue(len(expected_list), len(result_list))

        self.assertCountEqual(expected_list[0], result_list[0])
        self.assertCountEqual(expected_list[1], result_list[1])


if __name__ == "__main__":
    unittest.main()
